"""
Cement core interface module.
"""

from abc import ABC
from ..core import exc, meta
from ..utils.misc import minimal_logger

LOG = minimal_logger(__name__)


class Interface(ABC, meta.MetaMixin):

    """Base interface class that all Cement Interfaces should subclass from."""

    class Meta:

        """
        Interface meta-data (can also be passed as keyword arguments to the
        parent class).

        """

        interface = NotImplemented
        """The string identifier of this interface."""

    def __init__(self, **kw):
        super(Interface, self).__init__(**kw)
        try:
            assert self._meta.interface, \
                "%s.Meta.interface undefined." % self.__class__.__name__
        except AssertionError as e:
            raise(exc.InterfaceError(e.args[0]))

    def _validate(self):
        """
        Perform any validation to ensure proper data, meta-data, etc.
        """
        pass


class InterfaceManager(object):
    """
    Manages the interface system to define, get, list interfaces with
    the Cement Framework.

    """

    def __init__(self, app):
        self.app = app
        self.__interfaces__ = {}

    def get(self, interface, fallback=None, **kwargs):
        """
        Get an interface class.

        Args:
            interface (str): The interface label (i.e. ``output``)
            fallback (Handler):  A fallback value to return if ``interface``
                doesn't exist.

        Returns:
            Interface: An uninstantiated interface class

        Raises:
            cement.core.exc.InterfaceError: If the ``interface`` does not
                exist.

        Example:

            .. code-block:: python

                i = app.interface.get('output')

        """

        if interface in self.__interfaces__.keys():
            return self.__interfaces__[interface]
        elif fallback is not None:
            return fallback
        else:
            raise exc.InterfaceError("interface '%s' does not exist!" %
                                     interface)

    def list(self):
        """
        Return a list of defined interfaces.

        Returns:
            list: Interface labels.

        Example:

            .. code-block:: python

                app.interface.list()

        """
        return list(self.__interfaces__.keys())

    def define(self, ibc):
        """
        Define an ``ibc`` (interface base class).

        Args:
            ibc (Interface): The abstract base class that defines the interface

        Raises:
            cement.core.exc.InterfaceError: If the interface label is already
            defined

        Example:

            .. code-block:: python

                app.interface.define(DatabaseInterface)

        """

        LOG.debug("defining interface '%s' (%s)" %
                  (ibc.Meta.interface, ibc.__name__))

        if ibc.Meta.interface in self.__interfaces__:
            msg = "interface '%s' already defined!" % ibc.Meta.interface
            raise exc.InterfaceError(msg)
        self.__interfaces__[ibc.Meta.interface] = ibc

    def defined(self, interface):
        """
        Test whether ``interface`` is defined.

        Args:
            interface (str): The label of the interface (I.e.
                ``log``, ``config``, ``output``, etc).

        Returns:
            bool: ``True`` if the interface is defined, ``False`` otherwise

        Example:

            .. code-block:: python

                app.interface.defined('log')

        """
        if interface in self.__interfaces__:
            return True
        else:
            return False



"""
Cement plugin extension module.
"""

import os
import sys
import imp
import re
from ..core import plugin, exc
from ..utils.misc import is_true, minimal_logger
from ..utils.fs import abspath

LOG = minimal_logger(__name__)

# FIX ME: This is a redundant name... ?


class CementPluginHandler(plugin.PluginHandler):

    """
    This class is an internal implementation of the
    :ref:`IPlugin <cement.core.plugin>` interface. It does not take any
    parameters on initialization.

    """

    class Meta:

        """Handler meta-data."""

        label = 'cement'
        """The string identifier for this class."""

    def __init__(self):
        super().__init__()
        self._loaded_plugins = []
        self._enabled_plugins = []
        self._disabled_plugins = []

    def _setup(self, app_obj):
        super()._setup(app_obj)
        self._enabled_plugins = []
        self._disabled_plugins = []
        self.bootstrap = self.app._meta.plugin_module
        self.load_dirs = self.app._meta.plugin_dirs

        # parse all app configs for plugins. Note: these are already
        # loaded from files when app.config was setup.  The application
        # configuration OVERRIDES plugin configs.
        for section in self.app.config.get_sections():
            if not section.startswith('plugin.'):
                continue
            plugin_section = section
            plugin = re.sub('^plugin.', '', section)

            if 'enabled' not in self.app.config.keys(plugin_section):
                continue
            if is_true(self.app.config.get(plugin_section, 'enabled')):
                LOG.debug("enabling plugin '%s' per application config" %
                          plugin)
                if plugin not in self._enabled_plugins:
                    self._enabled_plugins.append(plugin)  # pragma: nocover
                if plugin in self._disabled_plugins:
                    self._disabled_plugins.remove(plugin)  # pragma: nocover
            else:
                LOG.debug("disabling plugin '%s' per application config" %
                          plugin)
                if plugin not in self._disabled_plugins:
                    self._disabled_plugins.append(plugin)  # pragma: nocover
                if plugin in self._enabled_plugins:
                    self._enabled_plugins.remove(plugin)  # pragma: nocover

    def _load_plugin_from_dir(self, plugin_name, plugin_dir):
        """
        Load a plugin from a directory path rather than a python package
        within sys.path.  This would either be ``myplugin.py`` or
        ``myplugin/__init__.py`` within the given ``plugin_dir``.

        Args:
            plugin_name (str): The name of the plugin.
            plugin_dir (str): The filesystem directory path where the plugin
                exists.

        """

        # FIX ME: `imp` is deprecated in Python 3.4 and will be  going away
        # so we need to update forward compatibility for ``importlib``.
        #
        # See: https://github.com/datafolklabs/cement/issues/386

        LOG.debug("attempting to load '%s' from '%s'" % (plugin_name,
                                                         plugin_dir))

        if not os.path.exists(plugin_dir):
            LOG.debug("plugin directory '%s' does not exist." % plugin_dir)
            return False

        try:
            f, path, desc = imp.find_module(plugin_name, [plugin_dir])
        except ImportError:
            LOG.debug("plugin '%s' does not exist in '%s'." %
                      (plugin_name, plugin_dir))
            return False

        # We don't catch this because it would make debugging a
        # nightmare
        mod = imp.load_module(plugin_name, f, path, desc)
        if mod and hasattr(mod, 'load'):
            mod.load(self.app)
        return True

    def _load_plugin_from_bootstrap(self, plugin_name, base_package):
        """
        Load a plugin from a python package.  Returns True if no ImportError
        is encountered.

        Args:
            plugin_name (str): The name of the plugin, also the name of the
                module to load from base_package. I.e.
                ``myapp.bootstrap.myplugin``
            base_package: The base python package to load the plugin module
                from.  I.e. ``myapp.bootstrap`` or similar.

        Returns:
            bool: ``True`` is the plugin was loaded, ``False`` otherwise

        Raises:
            :py:class:`ImportError`: If the plugin can not be imported

        """

        full_module = '%s.%s' % (base_package, plugin_name)

        # If the base package doesn't exist, we return False rather than
        # bombing out.
        if base_package not in sys.modules:
            try:
                __import__(base_package, globals(), locals(), [], 0)
            except ImportError:
                LOG.debug("unable to import plugin bootstrap module '%s'."
                          % base_package)
                return False

        LOG.debug("attempting to load '%s' from '%s'" % (plugin_name,
                                                         base_package))
        # We don't catch this because it would make debugging a nightmare
        # FIXME: not sure how to test/cover this
        if full_module not in sys.modules:
            __import__(full_module,
                       globals(), locals(), [], 0)  # pragma: nocover

        if hasattr(sys.modules[full_module], 'load'):
            sys.modules[full_module].load(self.app)

        return True

    def load_plugin(self, plugin_name):
        """
        Load a plugin whose name is ``plugin_name``.  First attempt to load
        from a plugin directory (plugin_dir), secondly attempt to load from a
        Python module determined by ``App.Meta.plugin_module``.

        Upon successful loading of a plugin, the plugin name is appended to
        the ``self._loaded_plugins list``.

        Args:
            plugin_name (str): The name of the plugin to load.

        Raises:
            cement.core.exc.FrameworkError: If the plugin can not be loaded

        """
        LOG.debug("loading application plugin '%s'" % plugin_name)

        # first attempt to load from plugin_dirs
        for load_dir in self.load_dirs:
            load_dir = abspath(load_dir)

            if self._load_plugin_from_dir(plugin_name, load_dir):
                self._loaded_plugins.append(plugin_name)
                break

        # then from a bootstrap module
        if plugin_name not in self._loaded_plugins:
            if self._load_plugin_from_bootstrap(plugin_name, self.bootstrap):
                self._loaded_plugins.append(plugin_name)

        # otherwise it's a bust
        if plugin_name not in self._loaded_plugins:
            raise exc.FrameworkError("Unable to load plugin '%s'." %
                                     plugin_name)

    def load_plugins(self, plugin_list):
        """
        Load a list of plugins.  Each plugin name is passed to
        ``self.load_plugin()``.

        Args:
            plugin_list (list): A list of plugin names to load.

        """
        for plugin_name in plugin_list:
            self.load_plugin(plugin_name)

    def get_loaded_plugins(self):
        """List of plugins that have been loaded."""
        return self._loaded_plugins

    def get_enabled_plugins(self):
        """List of plugins that are enabled (not necessary loaded yet)."""
        return self._enabled_plugins

    def get_disabled_plugins(self):
        """List of disabled plugins"""
        return self._disabled_plugins


def load(app):
    app.handler.register(CementPluginHandler)




from cement import minimal_logger

LOG = minimal_logger(__name__)

def {{ label }}_pre_run_hook(app):
    # do something with app
    LOG.debug('Inside {{ label }}_pre_run_hook!')

def load(app):
    # do something to extend cement
    app.hook.register('pre_run', {{ label }}_pre_run_hook)



"""
Cement tabulate extension module.
"""

from tabulate import tabulate
from ..core import output
from ..utils.misc import minimal_logger

LOG = minimal_logger(__name__)


class TabulateOutputHandler(output.OutputHandler):

    """
    This class implements the :ref:`Output <cement.core.output>` Handler
    interface.  It provides tabularized text output using the
    `Tabulate <https://pypi.python.org/pypi/tabulate>`_ module.  Please
    see the developer documentation on
    :cement:`Output Handling <dev/output>`.

    **Note** This extension has an external dependency on ``tabulate``.  You
    must include ``tabulate`` in your applications dependencies as Cement
    explicitly does **not** include external dependencies for optional
    extensions.
    """

    class Meta:

        """Handler meta-data."""

        label = 'tabulate'

        #: Whether or not to pad the output with an extra pre/post '\n'
        padding = True

        #: Default template format.  See the ``tabulate`` documentation for
        #: all supported template formats.
        format = 'orgtbl'

        #: Default headers to use.
        headers = []

        #: Default alignment for string columns.  See the ``tabulate``
        #: documentation for all supported ``stralign`` options.
        string_alignment = 'left'

        #: Default alignment for numeric columns.  See the ``tabulate``
        #: documentation for all supported ``numalign`` options.
        numeric_alignment = 'decimal'

        #: String format to use for float values.
        float_format = 'g'

        #: Default replacement for missing value.
        missing_value = ''

        #: Whether or not to include ``tabulate`` as an available to choice
        #: to override the ``output_handler`` via command line options.
        overridable = False

    def render(self, data, **kw):
        """
        Take a data dictionary and render it into a table.  Additional
        keyword arguments are passed directly to ``tabulate.tabulate``.

        Args:
            data_dict (dict): The data dictionary to render.

        Returns:
            str: The rendered template text

        """
        headers = kw.get('headers', self._meta.headers)

        out = tabulate(data, headers,
                       tablefmt=kw.get('tablefmt', self._meta.format),
                       stralign=kw.get(
                           'stralign', self._meta.string_alignment),
                       numalign=kw.get(
                           'numalign', self._meta.numeric_alignment),
                       missingval=kw.get(
                           'missingval', self._meta.missing_value),
                       floatfmt=kw.get('floatfmt', self._meta.float_format),
                       )
        out = out + '\n'

        if self._meta.padding is True:
            out = '\n' + out + '\n'

        return out


def load(app):
    app.handler.register(TabulateOutputHandler)



"""
Cement dummy extension module.
"""

from ..core.output import OutputHandler
from ..core.template import TemplateHandler
from ..core.mail import MailHandler
from ..utils.misc import minimal_logger

LOG = minimal_logger(__name__)


class DummyOutputHandler(OutputHandler):

    """
    This class is an internal implementation of the
    :class:`cement.core.output.OutputHandlerBase` interface. It does not take
    any parameters on initialization, and does not actually output anything.

    """
    class Meta:

        """Handler meta-data"""

        #: The string identifier of this handler.
        label = 'dummy'

        #: Whether or not to include ``dummy`` as an available to choice
        #: to override the ``output_handler`` via command line options.
        overridable = False

    def render(self, data, template=None, **kw):
        """
        This implementation does not actually render anything to output, but
        rather logs it to the debug facility.

        Args:
            data (dict): The data dictionary to render.

        Keyword Args:
            template (str): The template parameter is not used by this
                implementation at all.

        """
        LOG.debug("not rendering any output to console")
        LOG.debug("DATA: %s" % data)
        return None


class DummyTemplateHandler(TemplateHandler):

    """
    This class is an internal implementation of the
    :class:`cement.core.template.TemplateHandlerBase` interface. It does not
    take any parameters on initialization, and does not actually render
    anything.

    """
    class Meta:

        """Handler meta-data"""

        #: The string identifier of this handler.
        label = 'dummy'

    def render(self, content, data, *args, **kw):
        """
        This implementation does not actually render anything, but
        rather logs it to the debug facility.

        Args:
            content (str): The content to render as dictionary
            data (dict): The data dictionary to render.

        """
        LOG.debug("CONTENT: %s" % content)
        LOG.debug("DATA: %s" % data)
        return None

    def copy(self, src, dest, data):
        """
        This implementation does not actually copy anything, but rather logs it
        to the debug facility.

        Args:
            src (str): The source template directory.
            dest (str): The destination directory.
            data (dict): The data dictionary to render with templates.
        """
        LOG.debug("COPY: %s -> %s" % (src, dest))


class DummyMailHandler(MailHandler):

    """
    This class implements the :class:`cement.core.mail.IMail`
    interface, but is intended for use in development as no email is actually
    sent.

    Example:

    .. code-block:: python

        class MyApp(App):
            class Meta:
                label = 'myapp'
                mail_handler = 'dummy'

        with MyApp() as app:
            app.run()

            app.mail.send('This is my fake message',
                subject='This is my subject',
                to=['john@example.com', 'rita@example.com'],
                from_addr='me@example.com',
                )

    The above will print the following to console:

    .. code-block:: text

        ======================================================================
        DUMMY MAIL MESSAGE
        ----------------------------------------------------------------------

        To: john@example.com, rita@example.com
        From: me@example.com
        CC:
        BCC:
        Subject: This is my subject

        ---

        This is my fake message

        ----------------------------------------------------------------------

    **Configuration**

    This handler supports the following configuration settings:

     * **to** - Default ``to`` addresses (list, or comma separated depending
       on the ConfigHandler in use)
     * **from_addr** - Default ``from_addr`` address
     * **cc** - Default ``cc`` addresses (list, or comma separated depending
       on the ConfigHandler in use)
     * **bcc** - Default ``bcc`` addresses (list, or comma separated depending
       on the ConfigHandler in use)
     * **subject** - Default ``subject``
     * **subject_prefix** - Additional string to prepend to the ``subject``


    You can add these to any application configuration file under a
    ``[mail.dummy]`` section, for example:

    **~/.myapp.conf**

    .. code-block:: text

        [myapp]

        # set the mail handler to use
        mail_handler = dummy


        [mail.dummy]

        # default to addresses (comma separated list)
        to = me@example.com

        # default from address
        from = someone_else@example.com

        # default cc addresses (comma separated list)
        cc = jane@example.com, rita@example.com

        # default bcc addresses (comma separated list)
        bcc = blackhole@example.com, someone_else@example.com

        # default subject
        subject = This is The Default Subject

        # additional prefix to prepend to the subject
        subject_prefix = MY PREFIX >

    """

    class Meta:

        """Handler meta-data."""

        #: Unique identifier for this handler
        label = 'dummy'

    def _get_params(self, **kw):
        params = dict()
        for item in ['to', 'from_addr', 'cc', 'bcc', 'subject']:
            config_item = self.app.config.get(self._meta.config_section, item)
            params[item] = kw.get(item, config_item)

        # also grab the subject_prefix
        params['subject_prefix'] = self.app.config.get(
            self._meta.config_section,
            'subject_prefix'
        )

        return params

    def send(self, body, **kw):
        """
        Mimic sending an email message, but really just print what would be
        sent to console.  Keyword arguments override configuration
        defaults (cc, bcc, etc).

        Args:
            body: The message body to send

        Keyword Args:
            to (list): List of recipients (generally email addresses)
            from_addr (str): Address (generally email) of the sender
            cc (list): List of CC Recipients
            bcc (list): List of BCC Recipients
            subject (str): Message subject line

        Returns:
            bool:``True`` if message is sent successfully, ``False`` otherwise

        Example:

            .. code-block:: python

                # Using all configuration defaults
                app.mail.send('This is my message body')

                # Overriding configuration defaults
                app.mail.send('My message body'
                    to=['john@example.com'],
                    from_addr='me@example.com',
                    cc=['jane@example.com', 'rita@example.com'],
                    subject='This is my subject',
                    )

        """
        # shorted config values
        params = self._get_params(**kw)
        msg = "\n" + "=" * 77 + "\n"
        msg += "DUMMY MAIL MESSAGE\n"
        msg += "-" * 77 + "\n\n"
        msg += "To: %s\n" % ', '.join(params['to'])
        msg += "From: %s\n" % params['from_addr']
        msg += "CC: %s\n" % ', '.join(params['cc'])
        msg += "BCC: %s\n" % ', '.join(params['bcc'])

        if params['subject_prefix'] not in [None, '']:
            msg += "Subject: %s %s\n\n---\n\n" % (params['subject_prefix'],
                                                  params['subject'])
        else:
            msg += "Subject: %s\n\n---\n\n" % params['subject']
        msg += body + "\n"

        msg += "\n" + "-" * 77 + "\n"

        print(msg)
        return True


def load(app):
    app.handler.register(DummyOutputHandler)
    app.handler.register(DummyTemplateHandler)
    app.handler.register(DummyMailHandler)




import os
from .controllers.{{ label }} import {{ class_name }}

def add_template_dir(app):
    path = os.path.join(os.path.dirname(__file__), 'templates')
    app.add_template_dir(path)

def load(app):
    app.handler.register({{ class_name }})
    app.hook.register('post_setup', add_template_dir)




# flake8: noqa

from .core.foundation import App, TestApp
from .core.interface import Interface
from .core.handler import Handler
from .core.exc import FrameworkError, InterfaceError, CaughtSignal
from .ext.ext_argparse import ArgparseController as Controller
from .ext.ext_argparse import expose as ex
from .utils.misc import init_defaults, minimal_logger
from .utils import misc, fs, shell
from .utils.version import get_version



"""
Cement daemon extension module.
"""

import os
import sys
import io
import pwd
import grp
from ..core import exc
from ..utils.misc import minimal_logger

LOG = minimal_logger(__name__)
LOG = minimal_logger(__name__)
CEMENT_DAEMON_ENV = None
CEMENT_DAEMON_APP = None


class Environment(object):

    """
    This class provides a mechanism for altering the running processes
    environment.

    Optional Arguments:

    :keyword stdin: A file to read STDIN from.  Default: ``/dev/null``
    :keyword stdout: A file to write STDOUT to.  Default: ``/dev/null``
    :keyword stderr: A file to write STDERR to.  Default: ``/dev/null``
    :keyword dir: The directory to run the process in.
    :keyword pid_file: The filesystem path to where the PID (Process ID)
        should be written to.  Default: None
    :keyword user: The user name to run the process as.
        Default: ``os.getlogin()``
    :keyword group: The group name to run the process as.
        Default: The primary group of ``os.getlogin()``.
    :keyword umask: The umask to pass to os.umask().  Default: ``0``

    """

    def __init__(self, **kw):
        self.stdin = kw.get('stdin', '/dev/null')
        self.stdout = kw.get('stdout', '/dev/null')
        self.stderr = kw.get('stderr', '/dev/null')
        self.dir = kw.get('dir', os.curdir)
        self.pid_file = kw.get('pid_file', None)
        self.umask = kw.get('umask', 0)
        self.user = kw.get('user', pwd.getpwuid(os.getuid()).pw_name)

        # clean up
        self.dir = os.path.abspath(os.path.expanduser(self.dir))
        if self.pid_file:
            self.pid_file = os.path.abspath(os.path.expanduser(self.pid_file))

        try:
            self.user = pwd.getpwnam(self.user)
        except KeyError:
            raise exc.FrameworkError("Daemon user '%s' doesn't exist." %
                                     self.user)

        try:
            self.group = kw.get('group',
                                grp.getgrgid(self.user.pw_gid).gr_name)
            self.group = grp.getgrnam(self.group)
        except KeyError:
            raise exc.FrameworkError("Daemon group '%s' doesn't exist." %
                                     self.group)

    def _write_pid_file(self):
        """
        Writes ``os.getpid()`` out to ``self.pid_file``.
        """
        pid = str(os.getpid())
        LOG.debug('writing pid (%s) out to %s' % (pid, self.pid_file))

        # setup pid
        if self.pid_file:
            f = open(self.pid_file, 'w')
            f.write(pid)
            f.close()

            os.chown(self.pid_file, self.user.pw_uid, self.group.gr_gid)

    def switch(self):
        """
        Switch the current process's user/group to ``self.user``, and
        ``self.group``.  Change directory to ``self.dir``, and write the
        current pid out to ``self.pid_file``.
        """
        # set the running uid/gid
        LOG.debug('setting process uid(%s) and gid(%s)' %
                  (self.user.pw_uid, self.group.gr_gid))
        os.setgid(self.group.gr_gid)
        os.setuid(self.user.pw_uid)
        os.environ['HOME'] = self.user.pw_dir
        os.chdir(self.dir)
        if self.pid_file and os.path.exists(self.pid_file):
            raise exc.FrameworkError("Process already running (%s)" %
                                     self.pid_file)
        else:
            self._write_pid_file()

    def daemonize(self):  # pragma: no cover
        """
        Fork the current process into a daemon.

        References:

        UNIX Programming FAQ:
            1.7 How do I get my program to act like a daemon?
            http://www.unixguide.net/unix/programming/1.7.shtml
            http://www.faqs.org/faqs/unix-faq/programmer/faq/

        Advanced Programming in the Unix Environment
            W. Richard Stevens, 1992, Addison-Wesley, ISBN 0-201-56317-7.

        """
        LOG.debug('attempting to daemonize the current process')
        # Do first fork.
        try:
            pid = os.fork()
            if pid > 0:
                LOG.debug('successfully detached from first parent')
                os._exit(os.EX_OK)
        except OSError as e:
            sys.stderr.write("Fork #1 failed: (%d) %s\n" %
                             (e.errno, e.strerror))
            sys.exit(1)

        # Decouple from parent environment.
        os.chdir(self.dir)
        os.umask(int(self.umask))
        os.setsid()

        # Do second fork.
        try:
            pid = os.fork()
            if pid > 0:
                LOG.debug('successfully detached from second parent')
                os._exit(os.EX_OK)
        except OSError as e:
            sys.stderr.write("Fork #2 failed: (%d) %s\n" %
                             (e.errno, e.strerror))
            sys.exit(1)

        # Redirect standard file descriptors.
        stdin = open(self.stdin, 'r')
        stdout = open(self.stdout, 'a+')
        stderr = open(self.stderr, 'a+')

        if hasattr(sys.stdin, 'fileno'):
            try:
                os.dup2(stdin.fileno(), sys.stdin.fileno())
            except io.UnsupportedOperation:
                # FIXME: ?
                pass
        if hasattr(sys.stdout, 'fileno'):
            try:
                os.dup2(stdout.fileno(), sys.stdout.fileno())
            except io.UnsupportedOperation:
                # FIXME: ?
                pass
        if hasattr(sys.stderr, 'fileno'):
            try:
                os.dup2(stderr.fileno(), sys.stderr.fileno())
            except io.UnsupportedOperation:
                # FIXME: ?
                pass

        # Update our pid file
        self._write_pid_file()


def daemonize():  # pragma: no cover
    """
    This function switches the running user/group to that configured in
    ``config['daemon']['user']`` and ``config['daemon']['group']``.  The
    default user is ``os.getlogin()`` and the default group is that user's
    primary group.  A pid_file and directory to run in is also passed to the
    environment.

    It is important to note that with the daemon extension enabled, the
    environment will switch user/group/set pid/etc regardless of whether
    the ``--daemon`` option was passed at command line or not.  However, the
    process will only 'daemonize' if the option is passed to do so.  This
    allows the program to run exactly the same in forground or background.

    """
    # We want to honor the runtime user/group/etc even if --daemon is not
    # passed... but only daemonize if it is.
    global CEMENT_DAEMON_ENV
    global CEMENT_DAEMON_APP

    app = CEMENT_DAEMON_APP
    CEMENT_DAEMON_ENV = Environment(
        user=app.config.get('daemon', 'user'),
        group=app.config.get('daemon', 'group'),
        pid_file=app.config.get('daemon', 'pid_file'),
        dir=app.config.get('daemon', 'dir'),
        umask=app.config.get('daemon', 'umask'),
    )

    CEMENT_DAEMON_ENV.switch()

    # ugly, but backward compat since this was the way it was built and some
    # things might rely on calling app.daemonize() before app.run() is called
    if '--daemon' in app.argv or \
       hasattr(app, 'pargs') and app.pargs.daemon is True:
        CEMENT_DAEMON_ENV.daemonize()


def extend_app(app):
    """
    Adds the ``--daemon`` argument to the argument object, and sets the
    default ``[daemon]`` config section options.

    """
    global CEMENT_DAEMON_APP
    CEMENT_DAEMON_APP = app

    app.args.add_argument('--daemon', dest='daemon',
                          action='store_true', help='daemonize the process')

    # Add default config
    user = pwd.getpwuid(os.getuid())
    group = grp.getgrgid(user.pw_gid)

    defaults = dict()
    defaults['daemon'] = dict()
    defaults['daemon']['user'] = user.pw_name
    defaults['daemon']['group'] = group.gr_name
    defaults['daemon']['pid_file'] = None
    defaults['daemon']['dir'] = '/'
    defaults['daemon']['umask'] = 0
    app.config.merge(defaults, override=False)
    app.extend('daemonize', daemonize)


def cleanup(app):  # pragma: no cover
    """
    After application run time, this hook just attempts to clean up the
    pid_file if one was set, and exists.

    """
    global CEMENT_DAEMON_ENV

    if CEMENT_DAEMON_ENV and CEMENT_DAEMON_ENV.pid_file:
        if os.path.exists(CEMENT_DAEMON_ENV.pid_file):
            LOG.debug('Cleaning up pid_file...')
            pid = open(CEMENT_DAEMON_ENV.pid_file, 'r').read().strip()

            # only remove it if we created it.
            if int(pid) == int(os.getpid()):
                os.remove(CEMENT_DAEMON_ENV.pid_file)


def load(app):
    app.hook.register('post_setup', extend_app)
    app.hook.register('pre_close', cleanup)



"""
Cement colorlog extension module.
"""

import os
import sys
import logging
from colorlog import ColoredFormatter
from ..ext.ext_logging import LoggingLogHandler
from ..utils.misc import is_true


class ColorLogHandler(LoggingLogHandler):

    """
    This class implements the Log Handler interface.  It is
    a sub-class of :class:`cement.ext.ext_logging.LoggingLogHandler` which is
    based on the standard :py:class:`logging` library, and adds colorized
    console output using the
    `ColorLog <https://pypi.python.org/pypi/colorlog>`_ library.

    **Note** This extension has an external dependency on ``colorlog``.  You
    must include ``colorlog`` in your applications dependencies as Cement
    explicitly does **not** include external dependencies for optional
    extensions.

    """
    class Meta:

        """Handler meta-data."""

        #: The string identifier of the handler.
        label = "colorlog"

        #: Color mapping for each log level
        colors = {
            'DEBUG':    'cyan',
            'INFO':     'green',
            'WARNING':  'yellow',
            'ERROR':    'red',
            'CRITICAL': 'red,bg_white',
        }

        #: Default configuration settings.  Will be overridden by the same
        #: settings in any application configuration file under a
        #: ``[log.colorlog]`` block.
        config_defaults = dict(
            file=None,
            level='INFO',
            to_console=True,
            rotate=False,
            max_bytes=512000,
            max_files=4,
            colorize_file_log=False,
            colorize_console_log=True,
        )

        #: Formatter class to use for non-colorized logging (non-tty, file,
        #: etc)
        formatter_class_without_color = logging.Formatter

        #: Formatter class to use for colorized logging
        formatter_class = ColoredFormatter

    def _get_console_format(self):
        format = super(ColorLogHandler, self)._get_console_format()
        colorize = self.app.config.get('log.colorlog', 'colorize_console_log')
        if sys.stdout.isatty() or 'CEMENT_TEST' in os.environ:
            if is_true(colorize):
                format = "%(log_color)s" + format
        return format

    def _get_file_format(self):
        format = super(ColorLogHandler, self)._get_file_format()
        colorize = self.app.config.get('log.colorlog', 'colorize_file_log')
        if is_true(colorize):
            format = "%(log_color)s" + format
        return format

    def _get_console_formatter(self, format):
        colorize = self.app.config.get('log.colorlog', 'colorize_console_log')
        if sys.stdout.isatty() or 'CEMENT_TEST' in os.environ:
            if is_true(colorize):
                formatter = self._meta.formatter_class(
                    format,
                    log_colors=self._meta.colors
                )
            else:
                formatter = self._meta.formatter_class_without_color(
                    format
                )
        else:
            klass = self._meta.formatter_class_without_color  # pragma: nocover
            formatter = klass(format)                        # pragma: nocover

        return formatter

    def _get_file_formatter(self, format):
        colorize = self.app.config.get('log.colorlog', 'colorize_file_log')
        if is_true(colorize):
            formatter = self._meta.formatter_class(
                format,
                log_colors=self._meta.colors
            )
        else:
            formatter = self._meta.formatter_class_without_color(format)

        return formatter


def load(app):
    app.handler.register(ColorLogHandler)



"""Common File System Utilities."""

import os
import tempfile
import shutil


class Tmp(object):

    """
    Provides creation and cleanup of a separate temporary directory, and file.

    Keyword Arguments:
        cleanup (bool): Whether or not to delete the temporary directory and
            file on exit (when used with the ``with`` operator).
        suffix (str): The suffix that the directory and file will end with.
            Default: *no suffix*
        prefix (str): The prefix that the directory and file will start
            with. Default: *no prefix*
        dir (str): The parent directory path that the temp directory and file
            will be created in.  Default: *system default temporary path*

    Example:

        .. code-block:: python

            from cement.utils import fs

            with fs.Tmp() as tmp:
                # do something with a temporary directory
                os.path.listdir(tmp.dir)

                # do something with a temporary file
                with open(tmp.file, 'w') as f:
                    f.write('some data')

    """

    def __init__(self, **kwargs):
        self.cleanup = kwargs.get('cleanup', True)
        suffix = kwargs.get('suffix', '')
        prefix = kwargs.get('prefix', 'tmp')
        dir = kwargs.get('dir', None)

        self.dir = tempfile.mkdtemp(suffix=suffix,
                                    prefix=prefix,
                                    dir=dir)
        _, self.file = tempfile.mkstemp(suffix=suffix,
                                        prefix=prefix,
                                        dir=dir)

    def remove(self):
        """
        Remove the temporary directory (and file) if it exists, and
        ``self.cleanup`` is ``True``.
        """
        if self.cleanup is True:
            if os.path.exists(self.dir):
                shutil.rmtree(self.dir)
            if os.path.exists(self.file):
                os.remove(self.file)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.remove()


def abspath(path):
    """
    Return an absolute path, while also expanding the ``~`` user directory
    shortcut.

    Args:
        path (str): The original path to expand.

    Returns:
        str: The fully expanded, absolute path to the given ``path``

    Example:

        .. code-block:: python

            from cement.utils import fs

            fs.abspath('~/some/path')
            fs.abspath('./some.file')

    """
    return os.path.abspath(os.path.expanduser(path))


def join(*args, **kwargs):
    """
    Return a complete, joined path, by first calling ``abspath()`` on the first
    item to ensure the final path is complete.

    Args:
        paths (list): A list of paths to join together.

    Returns:
        list: The complete and absolute joined path.

    Example:

        .. code-block:: python

            from cement.utils import fs

            fs.join('~/some/path', 'some/other/relevant/paht')

    """
    paths = list(args)
    first_path = abspath(paths.pop(0))
    return os.path.join(first_path, *paths, **kwargs)


def join_exists(*paths):
    """
    Wrapper around ``os.path.join()``, ``os.path.abspath()``, and
    ``os.path.exists()``.

    Arguments:
        paths (list): List of paths to join, and then return ``True`` if that
                      path exists, or ``False`` if it does not.

    Returns:
        tuple: First item is the fully joined absolute path, and the second
               is ``bool`` (whether that path exists or not).
    """
    path = join(*paths)
    return (path, os.path.exists(path))


def ensure_dir_exists(path):
    """
    Ensure the directory ``path`` exists, and if not create it.

    Arguments:
        path (str): The filesystem path of a directory.

    Raises:
        AssertionError: If the directory ``path`` exists, but is not a
        directory.

    Returns: None
    """

    path = abspath(path)

    if os.path.exists(path) and not os.path.isdir(path):
        raise AssertionError('Path `%s` exists but is not a directory!' % path)
    elif not os.path.exists(path):
        os.makedirs(path)


def ensure_parent_dir_exists(path):
    """
    Ensure the parent directory of ``path`` (file, or directory) exists, and if
    not create it.

    Arguments:
        path (str): The filesystem path of a file or directory.

    Returns: None
    """

    parent_dir = os.path.dirname(abspath(path))
    return ensure_dir_exists(parent_dir)


def backup(path, suffix='.bak'):
    """
    Rename a file or directory safely without overwriting an existing
    backup of the same name.

    Args:
        path (str): The path to the file or directory to make a backup of.
        suffix (str): The suffix to rename files with.

    Returns:
        str: The new path of backed up file/directory

    Example:

        .. code-block:: python

            from cement.utils import fs

            fs.backup('/path/to/original/file')

    """
    count = -1
    new_path = None
    while True:
        if os.path.exists(path):
            if count == -1:
                new_path = "%s%s" % (path, suffix)
            else:
                new_path = "%s%s.%s" % (path, suffix, count)
            if os.path.exists(new_path):
                count += 1
                continue
            else:
                if os.path.isfile(path):
                    shutil.copy(path, new_path)
                elif os.path.isdir(path):
                    shutil.copytree(path, new_path)
                break
        else:
            break
    return new_path


# Kinda dirty, but should resolve issues on Windows per #183
if 'HOME' in os.environ:
    HOME_DIR = abspath(os.environ['HOME'])
else:
    HOME_DIR = abspath('~')  # pragma: nocover



"""
Cement jinja2 extension module.
"""

from ..core.output import OutputHandler
from ..core.template import TemplateHandler
from ..utils.misc import minimal_logger
from jinja2 import Environment, FileSystemLoader, PackageLoader

LOG = minimal_logger(__name__)


class Jinja2OutputHandler(OutputHandler):

    """
    This class implements the :ref:`OutputHandler <cement.core.output>`
    interface.  It provides text output from template and uses the
    `Jinja2 Templating Language
    <http://jinja.pocoo.org/>`_.
    Please see the developer documentation on
    :cement:`Output Handling <dev/output>`.

    """

    class Meta:

        """Handler meta-data."""

        label = 'jinja2'

    def __init__(self, *args, **kw):
        super(Jinja2OutputHandler, self).__init__(*args, **kw)
        self.templater = None

    def _setup(self, app):
        super(Jinja2OutputHandler, self)._setup(app)
        self.templater = self.app.handler.resolve('template', 'jinja2',
                                                  setup=True)

    def render(self, data, template=None, **kw):
        """
        Take a data dictionary and render it using the given template file.
        Additional keyword arguments are ignored.

        Args:
            data (dict): The data dictionary to render.

        Keyword Args:
            template (str): The path to the template, after the
                ``template_module`` or ``template_dirs`` prefix as defined in
                the application.

        Returns:
            str: The rendered template text

        """

        LOG.debug("rendering content using '%s' as a template." % template)
        content, _type, _path = self.templater.load(template)
        return self.templater.render(content, data)


class Jinja2TemplateHandler(TemplateHandler):

    """
    This class implements the :ref:`Template <cement.core.template>` Handler
    interface.  It renders content as template, and supports copying entire
    source template directories using the
    `Jinja2 Templating Language <http://jinja.pocoo.org/>`_.  Please
    see the developer documentation on
    :cement:`Template Handling <dev/template>`.

    **Note** This extension has an external dependency on ``jinja2``.  You
    must include ``jinja2`` in your applications dependencies as Cement
    explicitly does **not** include external dependencies for optional
    extensions.
    """

    class Meta:

        """Handler meta-data."""

        label = 'jinja2'

    def __init__(self, *args, **kw):
        super(Jinja2TemplateHandler, self).__init__(*args, **kw)

        # expose Jinja2 Environment instance so that we can manipulate it
        # higher in application code if necessary
        self.env = Environment(keep_trailing_newline=True)

    def load(self, *args, **kw):
        """
        Loads a template file first from ``self.app._meta.template_dirs`` and
        secondly from ``self.app._meta.template_module``.  The
        ``template_dirs`` have presedence.

        Args:
            template_path (str): The secondary path of the template **after**
                either ``template_module`` or ``template_dirs`` prefix (set via
                ``App.Meta``)

        Returns:
            tuple: The content of the template (``str``), the type of template
            (``str``: ``directory``, or ``module``), and the path (``str``) of
            the directory or module)

        Raises:
            cement.core.exc.FrameworkError: If the template does not exist in
                either the ``template_module`` or ``template_dirs``.
        """
        content, _type, _path = super(Jinja2TemplateHandler, self).load(*args,
                                                                        **kw)

        if _type == 'directory':
            self.env.loader = FileSystemLoader(self.app._meta.template_dirs)
        elif _type == 'module':
            parts = self.app._meta.template_module.rsplit('.', 1)
            self.env.loader = PackageLoader(parts[0], package_path=parts[1])

        return content, _type, _path

    def render(self, content, data, *args, **kw):
        """
        Render the given ``content`` as template with the ``data`` dictionary.

        Args:
            content (str): The template content to render.
            data (dict): The data dictionary to render.

        Returns:
            str: The rendered template text

        """
        LOG.debug("rendering content as text via %s" % self.__module__)

        if not isinstance(content, str):
            content = content.decode('utf-8')

        tmpl = self.env.from_string(content)
        res = tmpl.render(**data)
        return res


def load(app):
    app.handler.register(Jinja2OutputHandler)
    app.handler.register(Jinja2TemplateHandler)



"""
Cement yaml extension module.
"""

import yaml
from ..core import output
from ..utils.misc import minimal_logger
from ..ext.ext_configparser import ConfigParserConfigHandler

LOG = minimal_logger(__name__)


def suppress_output_before_run(app):
    """
    This is a ``post_argument_parsing`` hook that suppresses console output if
    the ``YamlOutputHandler`` is triggered via command line.

    :param app: The application object.

    """
    if not hasattr(app.pargs, 'output_handler_override'):
        return
    elif app.pargs.output_handler_override == 'yaml':
        app._suppress_output()


def unsuppress_output_before_render(app, data):
    """
    This is a ``pre_render`` that unsuppresses console output if
    the ``YamlOutputHandler`` is triggered via command line so that the Yaml
    is the only thing in the output.

    :param app: The application object.

    """
    if not hasattr(app.pargs, 'output_handler_override'):
        return
    elif app.pargs.output_handler_override == 'yaml':
        app._unsuppress_output()


def suppress_output_after_render(app, out_text):
    """
    This is a ``post_render`` hook that suppresses console output again after
    rendering, only if the ``YamlOutputHandler`` is triggered via command
    line.

    :param app: The application object.

    """
    if not hasattr(app.pargs, 'output_handler_override'):
        return
    elif app.pargs.output_handler_override == 'yaml':
        app._suppress_output()


class YamlOutputHandler(output.OutputHandler):

    """
    This class implements the :ref:`Output <cement.core.output>` Handler
    interface.  It provides Yaml output from a data dictionary and uses
    `pyYaml <http://pyYaml.org/wiki/PyYamlDocumentation>`_ to dump it to
    STDOUT.  Please see the developer documentation on
    :cement:`Output Handling <dev/output>`.

    This handler forces Cement to suppress console output until
    ``app.render`` is called (keeping the output pure Yaml).  If
    troubleshooting issues, you will need to pass the ``--debug`` option in
    order to unsuppress output and see what's happening.

    """

    class Meta:

        """Handler meta-data."""

        label = 'yaml'

        #: Whether or not to include ``yaml`` as an available choice
        #: to override the ``output_handler`` via command line options.
        overridable = False

    def __init__(self, *args, **kw):
        super(YamlOutputHandler, self).__init__(*args, **kw)
        self.config = None

    def _setup(self, app_obj):
        self.app = app_obj

    def render(self, data_dict, template=None, **kw):
        """
        Take a data dictionary and render it as Yaml output.  Note that the
        template option is received here per the interface, however this
        handler just ignores it.  Additional keyword arguments passed to
        ``yaml.dump()``.

        Args:
            data_dict (dict): The data dictionary to render.

        Keyword Args:
            template (str): Ignored in this output handler implementation.

        Returns:
            str: A Yaml encoded string.

        """
        LOG.debug("rendering output as yaml via %s" % self.__module__)
        return yaml.dump(data_dict, **kw)


class YamlConfigHandler(ConfigParserConfigHandler):

    """
    This class implements the :ref:`Config <cement.core.config>` Handler
    interface, and provides the same functionality of
    :ref:`ConfigParserConfigHandler <cement.ext.ext_configparser>`
    but with Yaml configuration files.  See
    `pyYaml <http://pyYaml.org/wiki/PyYamlDocumentation>`_ for more
    information on pyYaml.

    **Note** This extension has an external dependency on `pyYaml`.  You must
    include `pyYaml` in your application's dependencies as Cement explicitly
    does *not* include external dependencies for optional extensions.

    """
    class Meta:
        label = 'yaml'

    def __init__(self, *args, **kw):
        super(YamlConfigHandler, self).__init__(*args, **kw)

    def _parse_file(self, file_path):
        """
        Parse Yaml configuration file settings from file_path, overwriting
        existing config settings.  If the file does not exist, returns False.

        Args:
            file_path (str): The file system path to the Yaml configuration
                             file.

        """
        with open(file_path, 'r') as f:
            content = f.read()
            if content is not None and len(content) > 0:
                self.merge(yaml.load(content))

        return True


def load(app):
    app.hook.register('post_argument_parsing', suppress_output_before_run)
    app.hook.register('pre_render', unsuppress_output_before_render)
    app.hook.register('post_render', suppress_output_after_render)
    app.handler.register(YamlOutputHandler)
    app.handler.register(YamlConfigHandler)



"""Cement testing utilities."""

# flake8: noqa

import os
import shutil
from pytest import raises, skip
from ..core.foundation import TestApp
from ..core.exc import *
from ..utils.misc import rando
from ..utils import fs, shell



"""Cement core template module."""

import os
import sys
import pkgutil
import re
import shutil
from abc import abstractmethod
from ..core import exc
from ..core.interface import Interface
from ..core.handler import Handler
from ..utils.misc import minimal_logger
from ..utils import fs

LOG = minimal_logger(__name__)


class TemplateInterface(Interface):

    """
    This class defines the Template Interface.  Handlers that implement this
    interface must provide the methods and attributes defined below. In
    general, most implementations should sub-class from the provided
    :class:`TemplateHandler` base class as a starting point.
    """

    class Meta:

        """Handler meta-data."""

        #: The string identifier of the interface
        interface = 'template'

    @abstractmethod
    def render(self, content, data):
        """
        Render ``content`` as a template using the ``data`` dict.

        Args:
            content (str): The content to be rendered as a template.
            data (dict): The data dictionary to render with template.

        Returns:
            str: The rendered template string.

        """
        pass  # pragma: nocover

    @abstractmethod
    def copy(self, src, dest, data):
        """
        Render the ``src`` directory path, and copy to ``dest``.  This method
        must render directory and file **names** as template content, as well
        as the contents of files.

        Args:
            src (str): The source template directory path.
            dest (str): The destination directory path.
            data (dict): The data dictionary to render with template.
        Returns: None
        """
        pass  # pragma: nocover

    @abstractmethod
    def load(self, path):
        """
        Loads a template file first from ``self.app._meta.template_dirs`` and
        secondly from ``self.app._meta.template_module``.  The
        ``template_dirs`` have presedence.

        Args:
            path (str): The secondary path of the template **after**
                either ``template_module`` or ``template_dirs`` prefix (set via
                ``App.Meta``)

        Returns:
            tuple: The content of the template (``str``), the type of template
            (``str``: ``directory``, or ``module``), and the path (``str``) of
            the directory or module)

        Raises:
            cement.core.exc.FrameworkError: If the template does not exist in
                either the ``template_module`` or ``template_dirs``.
        """
        pass  # pragma: nocover


class TemplateHandler(TemplateInterface, Handler):
    """
    Base class that all template implementations should sub-class from.
    Keyword arguments passed to this class will override meta-data options.
    """

    class Meta:
        #: Unique identifier (str), used internally.
        label = None

        #: The interface that this handler implements.
        interface = 'template'

        #: List of file patterns to exclude (copy but not render as template)
        exclude = None

        #: List of file patterns to ignore completely (not copy at all)
        ignore = None

    def __init__(self, *args, **kwargs):
        super(TemplateHandler, self).__init__(*args, **kwargs)
        if self._meta.ignore is None:
            self._meta.ignore = []
        if self._meta.exclude is None:
            self._meta.exclude = []

    def render(self, content, data):
        """
        Render ``content`` as template using using the ``data`` dictionary.

        Args:
            content (str): The content to render.
            data (dict): The data dictionary to interpolate in the template.

        Returns:
            str: The rendered content.
        """

        # must be provided by a subclass
        raise NotImplementedError  # pragma: nocover

    def _match_patterns(self, item, patterns):
        for pattern in patterns:
            if re.match(pattern, item):
                return True
        return False

    def copy(self, src, dest, data, force=False, exclude=None, ignore=None):
        """
        Render ``src`` directory as template, including directory and file
        names, and copy to ``dest`` directory.

        Args:
            src (str): The source directory path.
            dest (str): The destination directory path.
            data (dict): The data dictionary to interpolate in the template.
            force (bool): Whether to overwrite existing files.
            exclude (list): List of regular expressions to match files that
                should only be copied, and not rendered as template.
            ignore (list): List of regular expressions to match files that
                should be completely ignored and not copied at all.

        Returns:
            bool: Returns ``True`` if the copy completed successfully.

        Raises:
            AssertionError: If the ``src`` template directory path does not
                exists, and when a ``dest`` file already exists and
                ``force is not True``.
        """

        dest = fs.abspath(dest)
        src = fs.abspath(src)
        escaped_src = src.encode('unicode-escape').decode('utf-8')

        # double escape for regex matching
        escaped_src_pattern = escaped_src.encode('unicode-escape')
        escaped_src_pattern = escaped_src_pattern.decode('utf-8')

        if exclude is None:
            exclude = []
        if ignore is None:
            ignore = []
        ignore_patterns = self._meta.ignore + ignore
        exclude_patterns = self._meta.exclude + exclude

        assert os.path.exists(src), "Source path %s does not exist!" % src

        if not os.path.exists(dest):
            os.makedirs(dest)

        LOG.debug('copying source template %s -> %s' % (src, dest))

        # here's the fun
        for cur_dir, sub_dirs, files in os.walk(src):
            escaped_cur_dir = cur_dir.encode('unicode-escape').decode('utf-8')

            if cur_dir == '.':
                continue    # pragma: nocover
            elif cur_dir == src:
                # don't render the source base dir (because we are telling it
                # where to go as `dest`)
                cur_dir_dest = dest
            elif self._match_patterns(cur_dir, ignore_patterns):
                LOG.debug(
                    'not copying ignored directory: %s' % cur_dir)
                continue
            elif self._match_patterns(cur_dir, exclude_patterns):
                LOG.debug(
                    'not rendering excluded directory as template: ' +
                    '%s' % cur_dir)

                cur_dir_stub = re.sub(escaped_src_pattern,
                                      '',
                                      escaped_cur_dir)
                cur_dir_stub = cur_dir_stub.lstrip('/')
                cur_dir_stub = cur_dir_stub.lstrip('\\\\')
                cur_dir_stub = cur_dir_stub.lstrip('\\')
                cur_dir_dest = os.path.join(dest, cur_dir_stub)
            else:
                # render the cur dir
                LOG.debug(
                    'rendering directory as template: %s' % cur_dir)

                cur_dir_stub = re.sub(escaped_src_pattern,
                                      '',
                                      escaped_cur_dir)
                cur_dir_stub = self.render(cur_dir_stub, data)
                cur_dir_stub = cur_dir_stub.lstrip('/')
                cur_dir_stub = cur_dir_stub.lstrip('\\\\')
                cur_dir_stub = cur_dir_stub.lstrip('\\')
                cur_dir_dest = os.path.join(dest, cur_dir_stub)

            # render sub-dirs
            for sub_dir in sub_dirs:
                escaped_sub_dir = sub_dir.encode('unicode-escape')
                escaped_sub_dir = escaped_sub_dir.decode('utf-8')

                full_path = os.path.join(cur_dir, sub_dir)

                if self._match_patterns(full_path, ignore_patterns):
                    LOG.debug(
                        'not copying ignored sub-directory: ' +
                        '%s' % full_path)
                    continue
                elif self._match_patterns(full_path, exclude_patterns):
                    LOG.debug(
                        'not rendering excluded sub-directory as template: ' +
                        '%s' % full_path)
                    sub_dir_dest = os.path.join(cur_dir_dest, sub_dir)
                else:
                    LOG.debug(
                        'rendering sub-directory as template: %s' % full_path)

                    new_sub_dir = re.sub(escaped_src_pattern,
                                         '',
                                         self.render(escaped_sub_dir, data))
                    sub_dir_dest = os.path.join(cur_dir_dest, new_sub_dir)

                if not os.path.exists(sub_dir_dest):
                    LOG.debug('creating sub-directory %s' % sub_dir_dest)
                    os.makedirs(sub_dir_dest)

            for _file in files:
                _rendered = self.render(_file, data)
                new_file = re.sub(escaped_src_pattern, '', _rendered)

                _file = fs.abspath(os.path.join(cur_dir, _file))
                _file_dest = fs.abspath(os.path.join(cur_dir_dest, new_file))

                # handle if destination path already exists

                if os.path.exists(_file_dest):
                    if force is True:
                        LOG.debug(
                            'overwriting existing file: %s ' % _file_dest)
                    else:
                        assert False, \
                            'Destination file already exists: %s ' % _file_dest

                if self._match_patterns(_file, ignore_patterns):
                    LOG.debug(
                        'not copying ignored file: ' +
                        '%s' % _file)
                    continue

                elif self._match_patterns(_file, exclude_patterns):
                    LOG.debug(
                        'not rendering excluded file: ' +
                        '%s' % _file)
                    shutil.copy(_file, _file_dest)

                else:
                    LOG.debug('rendering file as template: %s' % _file)
                    f = open(_file, 'r')
                    content = f.read()
                    f.close()

                    _file_content = self.render(content, data)
                    f = open(_file_dest, 'w')
                    f.write(_file_content)
                    f.close()

        return True

    def _load_template_from_file(self, template_path):
        for template_dir in self.app._meta.template_dirs:
            template_prefix = template_dir.rstrip('/')
            template_path = template_path.lstrip('/')
            full_path = fs.abspath(os.path.join(template_prefix,
                                                template_path))
            LOG.debug(
                "attemping to load output template from file %s" % full_path)
            if os.path.exists(full_path):
                content = open(full_path, 'r').read()
                LOG.debug("loaded output template from file %s" %
                          full_path)
                return (content, full_path)
            else:
                LOG.debug("output template file %s does not exist" %
                          full_path)
                continue

        return (None, None)

    def _load_template_from_module(self, template_path):
        template_module = self.app._meta.template_module
        template_path = template_path.lstrip('/')
        full_module_path = "%s.%s" % (template_module,
                                      re.sub('/', '.', template_path))

        LOG.debug("attemping to load output template '%s' from module %s" %
                  (template_path, template_module))

        # see if the module exists first
        if template_module not in sys.modules:
            try:
                __import__(template_module, globals(), locals(), [], 0)
            except ImportError:
                LOG.debug("unable to import template module '%s'."
                          % template_module)
                return (None, None)

        # get the template content
        try:
            content = pkgutil.get_data(template_module, template_path)
            LOG.debug("loaded output template '%s' from module %s" %
                      (template_path, template_module))
            return (content, full_module_path)
        except IOError:
            LOG.debug("output template '%s' does not exist in module %s" %
                      (template_path, template_module))
            return (None, None)

    def load(self, template_path):
        """
        Loads a template file first from ``self.app._meta.template_dirs`` and
        secondly from ``self.app._meta.template_module``.  The
        ``template_dirs`` have presedence.

        Args:
            template_path (str): The secondary path of the template **after**
                either ``template_module`` or ``template_dirs`` prefix (set via
                ``App.Meta``)

        Returns:
            tuple: The content of the template (``str``), the type of template
            (``str``: ``directory``, or ``module``), and the path (``str``) of
            the directory or module)

        Raises:
            cement.core.exc.FrameworkError: If the template does not exist in
                either the ``template_module`` or ``template_dirs``.
        """
        if not template_path:
            raise exc.FrameworkError("Invalid template path '%s'." %
                                     template_path)

        # first attempt to load from file
        content, path = self._load_template_from_file(template_path)
        if content is None:
            # second attempt to load from module
            content, path = self._load_template_from_module(template_path)
            template_type = 'module'
        else:
            template_type = 'directory'

        # if content is None, that means we didn't find a template file in
        # either and that is an exception
        if content is None:
            raise exc.FrameworkError("Could not locate template: %s" %
                                     template_path)

        return (content, template_type, path)



"""Cement core config module."""

import os
from abc import abstractmethod
from ..core.interface import Interface
from ..core.handler import Handler
from ..utils.fs import abspath
from ..utils.misc import minimal_logger

LOG = minimal_logger(__name__)


class ConfigInterface(Interface):

    """
    This class defines the Config Interface.  Handlers that implement this
    interface must provide the methods and attributes defined below. In
    general, most implementations should sub-class from the provided
    :class:`ConfigHandler` base class as a starting point.
    """

    class Meta:

        """Handler meta-data."""

        #: The string identifier of the interface.
        interface = 'config'

    @abstractmethod
    def parse_file(self, file_path):
        """
        Parse config file settings from ``file_path``.  Returns True if the
        file existed, and was parsed successfully.  Returns False otherwise.

        Args:
            file_path (str): The path to the config file to parse.

        Returns:
            bool: ``True`` if the file was parsed, ``False`` otherwise.

        """
        pass    # pragma: nocover

    @abstractmethod
    def keys(self, section):
        """
        Return a list of configuration keys from ``section``.

        Args:
            section (list): The config section to pull keys from.

        Returns:
            list: A list of keys in ``section``.

        """
        pass    # pragma: nocover

    @abstractmethod
    def get_sections(self):
        """
        Return a list of configuration sections.

        Returns:
            list: A list of config sections.

        """
        pass    # pragma: nocover

    @abstractmethod
    def get_dict(self):
        """
        Return a dict of the entire configuration.

        Returns:
            dict: A dictionary of the entire config.

        """

    @abstractmethod
    def get_section_dict(self, section):
        """
        Return a dict of configuration parameters for ``section``.

        Args:
            section (str): The config section to generate a dict from (using
                that sections' keys).

        Returns:
            dict: A dictionary of the config section.

        """
        pass    # pragma: nocover

    @abstractmethod
    def add_section(self, section):
        """
        Add a new section if it doesn't already exist.

        Args:
            section: The section label to create.

        Returns:
            None

        """
        pass    # pragma: nocover

    @abstractmethod
    def get(self, section, key):
        """
        Return a configuration value based on ``section.key``.  Must honor
        environment variables if they exist to override the config... for
        example ``config['myapp']['foo']['bar']`` must be overridable by the
        environment variable ``MYAPP_FOO_BAR``.... Note that ``MYAPP_`` must
        prefix all vars, therefore ``config['redis']['foo']`` would be
        overridable by ``MYAPP_REDIS_FOO`` ... but
        ``config['myapp']['foo']['bar']`` would not have a double prefix of
        ``MYAPP_MYAPP_FOO_BAR``.

        Args:
            section (str): The section of the configuration to pull key values
                from.
            key (str): The configuration key to get the value for.

        Returns:
            unknown: The value of the ``key`` in ``section``.

        """
        pass    # pragma: nocover

    @abstractmethod
    def set(self, section, key, value):
        """
        Set a configuration value based at ``section.key``.

        Args:
            section (str): The ``section`` of the configuration to pull key
                value from.
            key (str): The configuration key to set the value at.
            value: The value to set.

        Returns:
            None

        """
        pass    # pragma: nocover

    @abstractmethod
    def merge(self, dict_obj, override=True):
        """
        Merges a dict object into the configuration.

        Args:
            dict_obj (dict): The dictionary to merge into the config
            override (bool): Whether to override existing values or not.

        Returns:
            None

        """
        pass    # pragma: nocover

    @abstractmethod
    def has_section(self, section):
        """
        Returns whether or not the section exists.

        Args:
            section (str): The section to test for.

        Returns:
            bool: ``True`` if the configuration section exists, ``False``
                otherwise.

        """
        pass    # pragma: nocover


class ConfigHandler(ConfigInterface, Handler):

    """
    Config handler implementation.

    """

    @abstractmethod
    def _parse_file(self, file_path):
        """
        Parse a configuration file at ``file_path`` and store it.  This
        function must be provided by the handler implementation (that is
        sub-classing this).

        Args:
            file_path (str): The file system path to the configuration file.

        Returns:
            bool: ``True`` if file was read properly, ``False`` otherwise

        """
        pass    # pragma: nocover

    def parse_file(self, file_path):
        """
        Ensure we are using the absolute/expanded path to ``file_path``, and
        then call ``self._parse_file`` to parse config file settings from it,
        overwriting existing config settings.

        Developers sub-classing from here should generally override
        ``_parse_file`` which handles just the parsing of the file and leaving
        this function to wrap any checks/logging/etc.

        Args:
            file_path (str): The file system path to the configuration file.

        Returns:
            bool: ``True`` if the given ``file_path`` was parsed, and ``False``
                otherwise.

        """
        file_path = abspath(file_path)
        if os.path.exists(file_path):
            LOG.debug("config file '%s' exists, loading settings..." %
                      file_path)
            return self._parse_file(file_path)
        else:
            LOG.debug("config file '%s' does not exist, skipping..." %
                      file_path)
            return False




import os
import sys
sys.path.append(os.path.join(os.path.dirname(__file__), 'contrib'))

from cement import App, CaughtSignal  # noqa: E402
from .controllers.base import Base    # noqa: E402


class CementApp(App):
    class Meta:
        label = 'cement'
        controller = 'base'
        template_module = 'cement.cli.templates'
        template_handler = 'jinja2'
        config_handler = 'yaml'
        config_file_suffix = '.yml'

        extensions = [
            'generate',
            'yaml',
            'jinja2',
        ]

        handlers = [
            Base,
        ]


class CementTestApp(CementApp):
    class Meta:
        argv = []
        config_files = []
        exit_on_close = False


def main(argv=None):
    with CementApp() as app:
        try:
            app.run()
        except AssertionError as e:                     # pragma: nocover
            print('AssertionError > %s' % e.args[0])    # pragma: nocover
            app.exit_code = 1                           # pragma: nocover
        except CaughtSignal as e:                       # pragma: nocover
            print('\n%s' % e)                           # pragma: nocover
            app.exit_code = 0                           # pragma: nocover


if __name__ == '__main__':
    main()                                              # pragma: nocover



"""
Cement json extension module.
"""

from ..core import output
from ..utils.misc import minimal_logger
from ..ext.ext_configparser import ConfigParserConfigHandler

LOG = minimal_logger(__name__)


def suppress_output_before_run(app):
    """
    This is a ``post_argument_parsing`` hook that suppresses console output if
    the ``JsonOutputHandler`` is triggered via command line.

    :param app: The application object.

    """
    if not hasattr(app.pargs, 'output_handler_override'):
        return
    elif app.pargs.output_handler_override == 'json':
        app._suppress_output()


def unsuppress_output_before_render(app, data):
    """
    This is a ``pre_render`` that unsuppresses console output if
    the ``JsonOutputHandler`` is triggered via command line so that the JSON
    is the only thing in the output.

    :param app: The application object.

    """
    if not hasattr(app.pargs, 'output_handler_override'):
        return
    elif app.pargs.output_handler_override == 'json':
        app._unsuppress_output()


def suppress_output_after_render(app, out_text):
    """
    This is a ``post_render`` hook that suppresses console output again after
    rendering, only if the ``JsonOutputHandler`` is triggered via command
    line.

    :param app: The application object.

    """
    if not hasattr(app.pargs, 'output_handler_override'):
        return
    elif app.pargs.output_handler_override == 'json':
        app._suppress_output()


class JsonOutputHandler(output.OutputHandler):

    """
    This class implements the :ref:`Output <cement.core.output>` Handler
    interface.  It provides JSON output from a data dictionary using the
    `json <http://docs.python.org/library/json.html>`_ module of the standard
    library.  Please see the developer documentation on
    :cement:`Output Handling <dev/output>`.

    This handler forces Cement to suppress console output until
    ``app.render`` is called (keeping the output pure JSON).  If
    troubleshooting issues, you will need to pass the ``--debug`` option in
    order to unsuppress output and see what's happening.

    """
    class Meta:

        """Handler meta-data"""

        label = 'json'
        """The string identifier of this handler."""

        #: Whether or not to include ``json`` as an available choice
        #: to override the ``output_handler`` via command line options.
        overridable = False

        #: Backend JSON library module to use (`json`, `ujson`)
        json_module = 'json'

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self._json = None

    def _setup(self, app):
        super()._setup(app)
        self._json = __import__(self._meta.json_module,
                                globals(), locals(), [], 0)

    def render(self, data_dict, template=None, **kw):
        """
        Take a data dictionary and render it as Json output.  Note that the
        template option is received here per the interface, however this
        handler just ignores it.  Additional keyword arguments passed to
        ``json.dumps()``.

        Args:
            data_dict (dict): The data dictionary to render.

        Keyword Args:
            template: This option is completely ignored.

        Returns:
            str: A JSON encoded string.

        """
        LOG.debug("rendering output as Json via %s" % self.__module__)
        return self._json.dumps(data_dict, **kw)


class JsonConfigHandler(ConfigParserConfigHandler):

    """
    This class implements the :ref:`Config <cement.core.config>` Handler
    interface, and provides the same functionality of
    :ref:`ConfigParserConfigHandler <cement.ext.ext_configparser>`
    but with JSON configuration files.

    """
    class Meta:

        """Handler meta-data."""

        label = 'json'

        #: Backend JSON library module to use (`json`, `ujson`).
        json_module = 'json'

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self._json = None

    def _setup(self, app):
        super()._setup(app)
        self._json = __import__(self._meta.json_module,
                                globals(), locals(), [], 0)

    def _parse_file(self, file_path):
        """
        Parse JSON configuration file settings from file_path, overwriting
        existing config settings.  If the file does not exist, returns False.

        Args:
            file_path (str): The file system path to the JSON configuration
            file.

        Returns:
            bool

        """
        with open(file_path, 'r') as f:
            content = f.read()
            if content is not None and len(content) > 0:
                self.merge(self._json.loads(content))

        return True


def load(app):
    app.hook.register('post_argument_parsing', suppress_output_before_run)
    app.hook.register('pre_render', unsuppress_output_before_render)
    app.hook.register('post_render', suppress_output_after_render)
    app.handler.register(JsonOutputHandler)
    app.handler.register(JsonConfigHandler)



"""
Cement core log module.

"""

# from ..core import interface
from abc import abstractmethod
from ..core.interface import Interface
from ..core.handler import Handler


class LogInterface(Interface):

    """
    This class defines the Log Interface.  Handlers that implement this
    interface must provide the methods and attributes defined below. In
    general, most implementations should sub-class from the provided
    :class:`LogHandler` base class as a starting point.
    """

    class Meta:

        """Handler meta-data."""

        #: The string identifier of the interface.
        interface = 'log'

    @abstractmethod
    def set_level(self):
        """
        Set the log level.  Must except atleast one of:
            ``['INFO', 'WARNING', 'ERROR', 'DEBUG', or 'FATAL']``.

        """
        pass  # pragma: nocover

    @abstractmethod
    def get_level(self):
        """Return a string representation of the log level."""
        pass  # pragma: nocover

    @abstractmethod
    def info(self, msg):
        """
        Log to the ``INFO`` facility.

        Args:
            msg (str): The message to log.

        """
        pass  # pragma: nocover

    @abstractmethod
    def warning(self, msg):
        """
        Log to the ``WARNING`` facility.

        Args:
            msg (str): The message to log.

        """
        pass  # pragma: nocover

    @abstractmethod
    def error(self, msg):
        """
        Log to the ``ERROR`` facility.

        Args:
            msg (str): The message to log.

        """
        pass  # pragma: nocover

    @abstractmethod
    def fatal(self, msg):
        """
        Log to the ``FATAL`` facility.

        Args:
            msg (str): The message to log.

        """
        pass  # pragma: nocover

    @abstractmethod
    def debug(self, msg):
        """
        Log to the ``DEBUG`` facility.

        Args:
            msg (str): The message to log.

        """
        pass  # pragma: nocover


class LogHandler(LogInterface, Handler):

    """
    Log handler implementation.

    """

    pass  # pragma: nocover




from cement import App, CaughtSignal, Controller, ex, get_version

VERSION = (0, 0, 1, 'alpha', 0)

VERSION_BANNER = """
{{ label }} v%s
""" % get_version()


class Base(Controller):
    class Meta:
        label = 'base'

        arguments = [
            ### add a version banner
            ( [ '-v', '--version' ],
              { 'action'  : 'version',
                'version' : VERSION_BANNER } ),
        ]


    def _default(self):
        """Default action if no sub-command is passed."""

        self.app.args.print_help()


    @ex(
        help='example sub command1',
        arguments=[
            ### add a sample foo option under subcommand namespace
            ( [ '-f', '--foo' ],
              { 'help' : 'notorious foo option',
                'action'  : 'store',
                'dest' : 'foo' } ),
        ],
    )
    def command1(self):
        """Example sub-command."""

        print('Inside Base.command1')

        ### do something with arguments
        if self.app.pargs.foo is not None:
            print('Foo Argument > %s' % self.app.pargs.foo)


class MyApp(App):

    class Meta:
        # application label
        label = '{{ label }}'

        # register handlers
        handlers = [
            Base
        ]

        # call sys.exit() on close
        close_on_exit = True


def main():
    with MyApp() as app:
        try:
            app.run()
        except CaughtSignal as e:
            # Default Cement signals are SIGINT and SIGTERM, exit 0 (non-error)
            print('\n%s' % e)
            app.exit_code = 0


if __name__ == '__main__':
    main()




from pytest import raises
from {{ label }}.main import {{ class_name }}Test

def test_{{ label }}():
    # test {{ label }} without any subcommands or arguments
    with {{ class_name }}Test() as app:
        app.run()
        assert app.exit_code == 0


def test_{{ label }}_debug():
    # test that debug mode is functional
    argv = ['--debug']
    with {{ class_name }}Test(argv=argv) as app:
        app.run()
        assert app.debug is True


def test_command1():
    # test command1 without arguments
    argv = ['command1']
    with {{ class_name }}Test(argv=argv) as app:
        app.run()
        data,output = app.last_rendered
        assert data['foo'] == 'bar'
        assert output.find('Foo => bar')


    # test command1 with arguments
    argv = ['command1', '--foo', 'not-bar']
    with {{ class_name }}Test(argv=argv) as app:
        app.run()
        data,output = app.last_rendered
        assert data['foo'] == 'not-bar'
        assert output.find('Foo => not-bar')



"""
Cement mustache extension module.
"""

from pystache.renderer import Renderer
from ..core.output import OutputHandler
from ..core.template import TemplateHandler
from ..utils.misc import minimal_logger

LOG = minimal_logger(__name__)


class PartialsLoader(object):

    def __init__(self, handler):
        self.handler = handler

    def get(self, template):
        content, _type, _path = self.handler.load(template)
        return content


class MustacheOutputHandler(OutputHandler):

    """
    This class implements the :ref:`Output <cement.core.output>` Handler
    interface.  It provides text output from template and uses the
    `Mustache Templating Language <http://mustache.github.com>`_.  Please
    see the developer documentation on
    :cement:`Output Handling <dev/output>`.

    **Note** This extension has an external dependency on ``pystache``.  You
    must include ``pystache`` in your applications dependencies as Cement
    explicitly does **not** include external dependencies for optional
    extensions.
    """

    class Meta:

        """Handler meta-data."""

        label = 'mustache'

        #: Whether or not to include ``mustache`` as an available to choice
        #: to override the ``output_handler`` via command line options.
        overridable = False

    def __init__(self, *args, **kw):
        super(MustacheOutputHandler, self).__init__(*args, **kw)
        # self._partials_loader = PartialsLoader(self)
        self.templater = None

    def _setup(self, app):
        super(MustacheOutputHandler, self)._setup(app)
        self.templater = self.app.handler.resolve('template', 'mustache',
                                                  setup=True)

    def render(self, data, template=None, **kw):
        """
        Take a data dictionary and render it using the given template file.
        Additional keyword arguments passed to ``stache.render()``.

        Args:
            data (dict): The data dictionary to render.

        Keyword Args:
            template (str): The path to the template, after the
                ``template_module`` or ``template_dirs`` prefix as defined in
                the application.

        Returns:
            str: The rendered template text

        """

        LOG.debug("rendering content using '%s' as a template." % template)
        content, _type, _path = self.templater.load(template)
        return self.templater.render(content, data)


class MustacheTemplateHandler(TemplateHandler):

    """
    This class implements the :ref:`Template <cement.core.template>` Handler
    interface.  It renderd content as template, and supports copying entire
    source template directories using the
    `Mustache Templating Language <http://mustache.github.com>`_.  Please
    see the developer documentation on
    :cement:`Template Handling <dev/template>`.

    **Note** This extension has an external dependency on ``pystache``.  You
    must include ``pystache`` in your applications dependencies as Cement
    explicitly does **not** include external dependencies for optional
    extensions.
    """

    class Meta:

        """Handler meta-data."""

        label = 'mustache'

    def __init__(self, *args, **kw):
        super(MustacheTemplateHandler, self).__init__(*args, **kw)
        self._partials_loader = PartialsLoader(self)

    def render(self, content, data):
        """
        Render the given ``content`` as template with the ``data`` dictionary.

        Args:
            content (str): The template content to render.
            data (dict): The data dictionary to render.

        Returns:
            str: The rendered template text

        """

        stache = Renderer(partials=self._partials_loader)
        return stache.render(content, data)


def load(app):
    app.handler.register(MustacheOutputHandler)
    app.handler.register(MustacheTemplateHandler)



"""Cement core cache module."""

from abc import abstractmethod
from ..core.interface import Interface
from ..core.handler import Handler
from ..utils.misc import minimal_logger

LOG = minimal_logger(__name__)


class CacheInterface(Interface):

    """
    This class defines the Cache Interface.  Handlers that implement this
    interface must provide the methods and attributes defined below. In
    general, most implementations should sub-class from the provided
    :class:`CacheHandler` base class as a starting point.
    """

    class Meta:

        """Handler meta-data."""

        #: The string identifier of the interface.
        interface = 'cache'

    @abstractmethod
    def get(self, key, fallback=None):
        """
        Get the value for a key in the cache.

        If the key does not exist or the key/value in cache is expired, this
        functions must return ``fallback`` (which in turn must default to
        ``None``).

        Args:
            key (str): The key of the value stored in cache

        Keyword Args:
            fallback: Optional value that is returned if the cache is
                expired or the key does not exist.

        Returns:
            Unknown: Whatever the value is in the cache, or the ``fallback``

        """
        pass    # pragma: nocover

    @abstractmethod
    def set(self, key, value, time=None):
        """
        Set the key/value in the cache for a set amount of ``time``.

        Args:
            key (str): The key of the value to store in cache
            value (unknown): The value of that key to store in cache

        Keyword Args:
            time (int): A one-off expire time in seconds (or ``None``.  If no
                time is given, then a default value is used (determined by the
                implementation).

        Returns: None

        """
        pass    # pragma: nocover

    @abstractmethod
    def delete(self, key):
        """
        Deletes a key/value from the cache.

        Args:
            key: The key in the cache to delete

        Returns:
            bool: ``True`` if the key is successfully deleted, ``False``
            otherwise

        """
        pass    # pragma: nocover

    @abstractmethod
    def purge(self):
        """
        Clears all data from the cache.

        """
        pass    # pragma: nocover


class CacheHandler(CacheInterface, Handler):

    """
    Cache handler implementation.

    """
    pass    # pragma: nocover



"""Cement core controller module."""

from abc import abstractmethod
from ..core.interface import Interface
from ..core.handler import Handler
from ..utils.misc import minimal_logger

LOG = minimal_logger(__name__)


class ControllerInterface(Interface):

    """
    This class defines the Controller Interface.  Handlers that implement this
    interface must provide the methods and attributes defined below. In
    general, most implementations should sub-class from the provided
    :class:`ControllerHandler` base class as a starting point.
    """

    class Meta:

        """Interface meta-data."""

        #: The string identifier of the interface.
        interface = 'controller'

    @abstractmethod
    def _dispatch(self):
        """
        Reads the application object's data to dispatch a command from this
        controller.  For example, reading ``self.app.pargs`` to determine what
        command was passed, and then executing that command function.

        Note that Cement does *not* parse arguments when calling
        ``_dispatch()`` on a controller, as it expects the controller to
        handle parsing arguments (I.e. ``self.app.args.parse()``).

        Returns:
            unknown: The result of the executed controller function, or
            ``None`` if no controller function is called.

        """
        pass    # pragma: nocover


class ControllerHandler(ControllerInterface, Handler):
    """Controller handler implementation."""
    pass    # pragma: nocover



"""
Cement configparser extension module.
"""

import os
import re
from ..core import config
from ..utils.misc import minimal_logger
from configparser import RawConfigParser

LOG = minimal_logger(__name__)


class ConfigParserConfigHandler(config.ConfigHandler, RawConfigParser):

    """
    This class is an implementation of the :ref:`Config <cement.core.config>`
    interface.  It handles configuration file parsing and the like by
    sub-classing from the standard `ConfigParser
    <http://docs.python.org/library/configparser.html>`_
    library.  Please see the ConfigParser documentation for full usage of the
    class.

    Additional arguments and keyword arguments are passed directly to
    RawConfigParser on initialization.
    """
    class Meta:

        """Handler meta-data."""

        label = 'configparser'
        """The string identifier of this handler."""

    def merge(self, dict_obj, override=True):
        """
        Merge a dictionary into our config.  If override is True then
        existing config values are overridden by those passed in.

        Args:
            dict_obj (dict): A dictionary of configuration keys/values to merge
                into our existing config (self).

        Keyword Args:
            override (bool):  Whether or not to override existing values in the
                config.

        """
        assert isinstance(dict_obj, dict), "Dictionary object required."

        for section in list(dict_obj.keys()):
            if type(dict_obj[section]) == dict:
                if section not in self.get_sections():
                    self.add_section(section)

                for key in list(dict_obj[section].keys()):
                    if override:
                        self.set(section, key, dict_obj[section][key])
                    else:
                        # only set it if the key doesn't exist
                        if key not in self.keys(section):
                            self.set(section, key, dict_obj[section][key])

                # we don't support nested config blocks, so no need to go
                # further down to more nested dicts.

    def _parse_file(self, file_path):
        """
        Parse a configuration file at ``file_path`` and store it.

        Args:
            file_path (str): The file system path to the configuration file.

        Returns:
            bool: ``True`` if file was read properly, ``False`` otherwise

        """
        self.read(file_path)

        # FIX ME: Should check that file was read properly, however if not it
        # will likely raise an exception anyhow.
        return True

    def keys(self, section):
        """
        Return a list of keys within ``section``.

        Args:
            section (str): The config section

        Returns:
            list: List of keys in the ``section``.

        """
        return self.options(section)

    def get_dict(self):
        """
        Return a dict of the entire configuration.

        Returns:
            dict: A dictionary of the entire config.
        """
        _config = {}
        for section in self.get_sections():
            _config[section] = self.get_section_dict(section)
        return _config

    def get_sections(self):
        """
        Return a list of configuration sections.

        Returns:
            list: List of sections

        """
        return self.sections()

    def get_section_dict(self, section):
        """
        Return a dict representation of a section.

        Args:
            section: The section of the configuration.

        Returns:
            dict: Dictionary reprisentation of the config section.

        """
        dict_obj = dict()
        for key in self.keys(section):
            dict_obj[key] = self.get(section, key)
        return dict_obj

    def add_section(self, section):
        """
        Adds a block section to the config.

        Args:
            section (str): The section to add.

        """
        return RawConfigParser.add_section(self, section)

    def _get_env_var(self, section, key):
        if section == self.app._meta.config_section:
            env_var = "%s_%s" % (self.app._meta.config_section, key)
        else:
            env_var = "%s_%s_%s" % (
                self.app._meta.config_section, section, key)

        env_var = env_var.upper()
        env_var = re.sub('[^0-9a-zA-Z]+', '_', env_var)
        return env_var

    def get(self, section, key):
        env_var = self._get_env_var(section, key)
        if env_var in os.environ.keys():
            return os.environ[env_var]
        else:
            return RawConfigParser.get(self, section, key)

    def has_section(self, section):
        return RawConfigParser.has_section(self, section)

    def set(self, section, key, value):
        return RawConfigParser.set(self, section, key, value)


def load(app):
    app.handler.register(ConfigParserConfigHandler)



"""Cement core mail module."""

from abc import abstractmethod
from ..core.interface import Interface
from ..core.handler import Handler
from ..utils.misc import minimal_logger

LOG = minimal_logger(__name__)


class MailInterface(Interface):

    """
    This class defines the Mail Interface.  Handlers that implement this
    interface must provide the methods and attributes defined below. In
    general, most implementations should sub-class from the provided
    :class:`MailHandler` base class as a starting point.
    """

    class Meta:

        """Handler meta-data."""

        interface = 'mail'
        """The label identifier of the interface."""

    @abstractmethod
    def send(self, body, **kwargs):
        """
        Send a mail message.  Keyword arguments override configuration
        defaults (cc, bcc, etc).

        Args:
            body (str): The message body to send

        Keyword Args:
            to (list): List of recipients (generally email addresses)
            from_addr (str): Address (generally email) of the sender
            cc (list): List of CC Recipients
            bcc (list): List of BCC Recipients
            subject (str): Message subject line

        Returns:
            bool: ``True`` if message was sent successfully, ``False``
            otherwise

        Example:

            .. code-block:: python

                # Using all configuration defaults
                app.mail.send('This is my message body')

                # Overriding configuration defaults
                app.mail.send('My message body'
                    to=['john@example.com'],
                    from_addr='me@example.com',
                    cc=['jane@example.com', 'rita@example.com'],
                    subject='This is my subject',
                    )

        """
        pass  # pragma: nocover


class MailHandler(MailInterface, Handler):

    """
    Mail handler implementation.

    **Configuration**

    This handler supports the following configuration settings:

     * **to** - Default ``to`` addresses (list, or comma separated depending
       on the ConfigHandler in use)
     * **from_addr** - Default ``from_addr`` address
     * **cc** - Default ``cc`` addresses (list, or comma separated depending
       on the ConfigHandler in use)
     * **bcc** - Default ``bcc`` addresses (list, or comma separated depending
       on the ConfigHandler in use)
     * **subject** - Default ``subject``
     * **subject_prefix** - Additional string to prepend to the ``subject``

    """

    class Meta:

        """
        Handler meta-data (can be passed as keyword arguments to the parent
        class).
        """

        #: Configuration default values
        config_defaults = {
            'to': [],
            'from_addr': 'noreply@example.com',
            'cc': [],
            'bcc': [],
            'subject': 'Default Subject Line',
            'subject_prefix': '',
        }

    def _setup(self, app_obj):
        super()._setup(app_obj)
        self._validate_config()

    def _validate_config(self):
        # convert comma separated strings to lists (ConfigParser)
        for item in ['to', 'cc', 'bcc']:
            if item in self.app.config.keys(self._meta.config_section):
                value = self.app.config.get(self._meta.config_section, item)

                # convert a comma-separated string to a list
                if type(value) is str:
                    value_list = value.split(',')

                    # clean up extra space if they had it inbetween commas
                    value_list = [x.strip() for x in value_list]

                    # set the new extensions value in the config
                    self.app.config.set(self._meta.config_section, item,
                                        value_list)



"""Cement core foundation module."""

import os
import sys
import signal
import platform
from time import sleep
from ..core import exc, log, config, plugin
from ..core import output, extension, arg, controller, meta, cache, mail
from ..core import template
from ..core.interface import InterfaceManager
from ..core.handler import HandlerManager
from ..core.hook import HookManager
from ..utils.misc import is_true, minimal_logger
from ..utils import fs, misc
from ..ext.ext_argparse import ArgparseController as Controller

join = os.path.join

# The `imp` module is deprecated in favor of `importlib` in 3.4, but it
# wasn't introduced until 3.1.  Finally, reload is a builtin on Python < 3
pyver = sys.version_info
if pyver[0] >= 3 and pyver[1] >= 4:                # pragma: nocover  # noqa
    from importlib import reload as reload_module  # pragma: nocover  # noqa
elif pyver[0] >= 3:                                # pragma: nocover  # noqa
    from imp import reload as reload_module        # pragma: nocover  # noqa
else:                                              # pragma: nocover  # noqa
    reload_module = reload                         # pragma: nocover  # noqa


LOG = minimal_logger(__name__)
if platform.system() == 'Windows':
    SIGNALS = [signal.SIGTERM, signal.SIGINT]   # pragma: nocover
else:
    SIGNALS = [signal.SIGTERM, signal.SIGINT, signal.SIGHUP]


def add_handler_override_options(app):
    """
    This is a ``post_setup`` hook that adds the handler override options to
    the argument parser

    Args:
        app (instance): The application object

    """
    if app._meta.handler_override_options is None:
        return

    for i in app._meta.handler_override_options:
        if i not in app.interface.list():
            LOG.debug("interface '%s'" % i +
                      " is not defined, can not override handlers")
            continue

        if len(app.handler.list(i)) > 1:
            handlers = []
            for h in app.handler.list(i):
                handlers.append(app._resolve_handler(i, h))

            choices = [x._meta.label
                       for x in handlers
                       if x._meta.overridable is True]

            # don't display the option if no handlers are overridable
            if not len(choices) > 0:
                LOG.debug("no handlers are overridable within the " +
                          "%s interface" % i)
                continue

            # override things that we need to control
            argument_kw = app._meta.handler_override_options[i][1]
            argument_kw['dest'] = '%s_handler_override' % i
            argument_kw['action'] = 'store'
            argument_kw['choices'] = choices

            app.args.add_argument(
                *app._meta.handler_override_options[i][0],
                **app._meta.handler_override_options[i][1]
            )


def handler_override(app):
    """
    This is a ``post_argument_parsing`` hook that overrides a configured
    handler if defined in ``App.Meta.handler_override_options`` and
    the option is passed at command line with a valid handler label.

    Args:
        app (instance): The application object.

    """
    if app._meta.handler_override_options is None:
        return

    for i in app._meta.handler_override_options.keys():
        if not hasattr(app.pargs, '%s_handler_override' % i):
            continue  # pragma: nocover
        elif getattr(app.pargs, '%s_handler_override' % i) is None:
            continue  # pragma: nocover
        else:
            # get the argument value from command line
            argument = getattr(app.pargs, '%s_handler_override' % i)
            setattr(app._meta, '%s_handler' % i, argument)

            # and then re-setup the handler
            getattr(app, '_setup_%s_handler' % i)()


def cement_signal_handler(signum, frame):
    """
    Catch a signal, run the ``signal`` hook, and then raise an exception
    allowing the app to handle logic elsewhere.

    Args:
        signum (int): The signal number
        frame: The signal frame

    Raises:
        cement.core.exc.CaughtSignal: Raised, passing ``signum``, and ``frame``

    """
    LOG.debug('Caught signal %s' % signum)

    # FIXME: Maybe this isn't ideal... purhaps make
    # App.Meta.signal_handler a decorator that take the app object
    # and wraps/returns the actually signal handler?
    for f_global in frame.f_globals.values():
        if isinstance(f_global, App):
            app = f_global
            for res in app.hook.run('signal', app, signum, frame):
                pass  # pragma: nocover
    raise exc.CaughtSignal(signum, frame)


class App(meta.MetaMixin):

    """The primary application object class."""

    class Meta:

        """
        Application meta-data (can also be passed as keyword arguments to the
        parent class).
        """

        label = None
        """
        The name of the application.  This should be the common name as you
        would see and use at the command line.  For example ``helloworld``, or
        ``my-awesome-app``.
        """

        debug = False
        """
        Used internally, and should not be used by developers.  This is set
        to ``True`` if the ``debug`` option is passed at command line."""

        debug_argument_options = ['-d', '--debug']
        """
        The argument option(s) to toggle debug mode via cli.
        """

        debug_argument_help = 'full application debug mode'
        """
        The debug argument help text that is displayed in ``--help``.
        """

        quiet = False
        """
        Used internally, and should not be used by developers.  This is set
        to ``True`` if the ``quiet`` option is passed at command line."""

        quiet_argument_options = ['-q', '--quiet']
        """
        The argument option(s) to toggle quiet mode via cli.
        """

        quiet_argument_help = 'suppress all console output'
        """
        The quiet argument help text that is displayed in ``--help``.
        """

        exit_on_close = False
        """
        Whether or not to call ``sys.exit()`` when ``close()`` is called.
        The default is ``False``, however if ``True`` then the app will call
        ``sys.exit(X)`` where ``X`` is ``self.exit_code``.
        """

        config_file_suffix = '.conf'
        """
        Extension used to identify application and plugin configuration files.
        """

        config_files = None
        """
        List of config files to parse (appended to the builtin list of config
        files defined by Cement).

        Note: Though ``App.Meta.config_files`` defaults to ``None``, Cement
        will set this to a default list based on ``App.Meta.label`` (or in
        other words, the name of the application).  This will equate to:

        .. code-block:: python

            [
                '/etc/myapp/myapp.conf',
                '~/.config/myapp/myapp.conf',
                '~/.myapp.conf',
            ]


        Files are loaded in order, and have precedence in order.  Therefore,
        the last configuration loaded has precedence (and overwrites settings
        loaded from previous configuration files).

        Note that ``.conf`` is the default config file extension, defined by
        ``App.Meta.config_file_suffix``.
        """

        config_dirs = None
        """
        List of config directories to search config files (appended to the
        builtin list of directories defined by Cement). For each directory
        cement will load all files that ends with ``.conf``. Note: Though
        ``App.Meta.config_dirs`` defaults to ``None``, Cement
        will set this to a default list based on ``App.Meta.label`` (or in
        other words, the name of the application).  This will equate to:

        .. code-block:: python

            [
                '/etc/myapp/ext.d/',
                '/etc/myapp/plugins.d/',
                '~/.myapp/config/ext.d/',
                '~/.myapp/config/plugins.d/',
            ]

        Directories and files inside are loaded in order, and have precedence
        in order.  Therefore, the last configuration loaded has precedence
        (and overwrites settings loaded from previous configuration files).
        These configuration will be overriden by configuration from
        ``CementApp.Meta.config_files``.

        Note that ``.conf`` is the default config file extension, defined by
        ``CementApp.Meta.config_file_suffix``.
        """

        plugins = []
        """
        A list of plugins to load.  This is generally considered bad practice
        since plugins should be dynamically enabled/disabled via a plugin
        config file.
        """

        plugin_module = None
        """
        A python package (dotted import path) where plugin code can be
        loaded from.  This is generally something like ``myapp.plugins``
        where a plugin file would live at ``myapp/plugins/myplugin.py`` or
        ``myapp/plugins/myplugin/`` (directory). This provides a facility for
        applications that have builtin plugins that ship with the applications
        source code and live in the same Python module.

        Note: Though the meta default is ``None``, Cement will set this to
        ``<app_label>.plugins`` if not set.
        """

        plugin_dirs = None
        """
        A list of directory paths where plugin code (modules) can be loaded
        from (appended to the builtin list of directories defined by Cement).

        Note: Though ``App.Meta.plugin_dirs`` defaults to ``None``, Cement will
        populate this with a default list based on ``App.Meta.label``.
        This will equate to:

        .. code-block:: python

            [
                '~/.myapp/plugins',
                '~/.config/myapp/plugins',
                '/usr/lib/myapp/plugins',
            ]


        Modules are attempted to be loaded in order, and will stop loading
        once a plugin is successfully loaded from a directory.  Therefore
        this is the oposite of configuration file loading, in that here the
        first has precedence.
        """

        plugin_dir = None
        """
        A directory path where plugin code (modules) can be loaded from.
        By default, this setting is also overridden by the
        ``myapp.plugin_dir`` config setting parsed in any of the
        application configuration files.

        If set, this item will be **prepended** to ``Meta.plugin_dirs`` so
        that a users defined ``plugin_dir`` has precedence over others.

        In general, this setting should not be defined by the developer, as it
        is primarily used to allow the end-user to define a
        ``plugin_dir`` without completely trumping the hard-coded list
        of default ``plugin_dirs`` defined by the app/developer.
        """

        argv = None
        """
        A list of arguments to use for parsing command line arguments
        and options.

        Note: Though ``App.Meta.argv`` defaults to ``None``, Cement will set
        this to ``list(sys.argv[1:])`` if no argv is set in Meta during
        ``setup()``.
        """

        core_handler_override_options = dict(
            output=(['-o'], dict(help='output handler')),
        )
        """
        Similar to ``App.Meta.handler_override_options`` but these are
        the core defaults required by Cement.  This dictionary can be
        overridden by ``App.Meta.handler_override_options`` (when they
        are merged together).
        """

        handler_override_options = {}
        """
        Dictionary of handler override options that will be added to the
        argument parser, and allow the end-user to override handlers.  Useful
        for interfaces that have multiple uses within the same application
        (for example: Output Handler (json, yaml, etc) or maybe a Cloud
        Provider Handler (rackspace, digitalocean, amazon, etc).

        This dictionary will merge with
        ``App.Meta.core_handler_override_options`` but this one has
        precedence.

        Dictionary Format:

        .. code-block:: text

            <interface_name> = (option_arguments, help_text)


        See ``App.Meta.core_handler_override_options`` for an example
        of what this should look like.

        Note, if set to ``None`` then no options will be defined, and the
        ``App.Meta.core_meta_override_options`` will be ignore (not
        recommended as some extensions rely on this feature).
        """

        config_section = None
        """
        The base configuration section for the application.

        Note: Though ``App.Meta.config_section`` defaults to ``None``, Cement
        will set this to the value of ``App.Meta.label`` (or in other words,
        the name of the application).
        """

        config_defaults = None
        """Default configuration dictionary.  Must be of type ``dict``."""

        meta_defaults = {}
        """
        Default meta-data dictionary used to pass high level options from the
        application down to handlers at the point they are registered by the
        framework **if the handler has not already been instantiated**.

        For example, if requiring the ``json`` extension, you might want to
        override ``JsonOutputHandler.Meta.json_module`` with ``ujson`` by
        doing the following:

        .. code-block:: python

            from cement import App

            META = {
                'output.json': {
                    'json_module': 'ujson',
                }
            }

            class MyApp(App):
                class Meta:
                    label = 'myapp'
                    extensions = ['json']
                    meta_defaults = META

        """

        catch_signals = SIGNALS
        """
        List of signals to catch, and raise ``cement.core.exc.CaughtSignal``
        for.  Can be set to ``None`` to disable signal handling.
        """

        signal_handler = cement_signal_handler
        """A function that is called to handle any caught signals."""

        config_handler = 'configparser'
        """
        A handler class that implements the Config interface.
        """

        mail_handler = 'dummy'
        """
        A handler class that implements the Mail interface.
        """

        extension_handler = 'cement'
        """
        A handler class that implements the Extension interface.
        """

        log_handler = 'logging'
        """
        A handler class that implements the Log interface.
        """

        plugin_handler = 'cement'
        """
        A handler class that implements the Plugin interface.
        """

        argument_handler = 'argparse'
        """
        A handler class that implements the Argument interface.
        """

        output_handler = 'dummy'
        """
        A handler class that implements the Output interface.
        """

        template_handler = 'dummy'
        """
        A handler class that implements the Template interface.
        """

        cache_handler = None
        """
        A handler class that implements the Cache interface.
        """

        extensions = []
        """List of additional framework extensions to load."""

        bootstrap = None
        """
        A bootstrapping module to load after app creation, and before
        ``app.setup()`` is called.  This is useful for larger applications
        that need to offload their bootstrapping code such as registering
        hooks/handlers/etc to another file.

        This must be a dotted python module path.
        I.e. ``myapp.bootstrap`` (``myapp/bootstrap.py``).  Cement will then
        import the module, and if the module has a ``load()`` function, that
        will also be called.  Essentially, this is the same as an
        extension or plugin, but as a facility for the application itself
        to bootstrap hard-coded application code.  It is also called
        before plugins are loaded.
        """

        core_extensions = [
            'cement.ext.ext_dummy',
            'cement.ext.ext_smtp',
            'cement.ext.ext_plugin',
            'cement.ext.ext_configparser',
            'cement.ext.ext_logging',
            'cement.ext.ext_argparse',
        ]
        """
        List of Cement core extensions.  These are generally required by
        Cement and should only be modified if you know what you're
        doing.  Use ``App.Meta.extensions`` to add to this list, rather than
        overriding core extensions.  That said if you want to prune down
        your application, you can remove core extensions if they are
        not necessary (for example if using your own log handler
        extension you might not need/want ``LoggingLogHandler`` to be
        registered).
        """

        core_meta_override = [
            'debug',
            # 'plugin_config_dir',
            'plugin_dir',
            'ignore_deprecation_warnings',
            'template_dir',
            'mail_handler',
            'cache_handler',
            'log_handler',
            'output_handler',
            'template_handler',
        ]
        """
        List of meta options that can/will be overridden by config options
        of the ``base`` config section (where ``base`` is the base
        configuration section of the application which is determined by
        ``App.Meta.config_section`` but defaults to ``App.Meta.label``). These
        overrides are required by the framework to function properly and should
        not be used by end-user (developers) unless you really know what
        you're doing.  To add your own extended meta overrides you should use
        ``App.Meta.meta_override``.
        """

        meta_override = []
        """
        List of meta options that can/will be overridden by config options
        of the ``base`` config section (where ``base`` is the
        base configuration section of the application which is determined
        by ``App.Meta.config_section`` but defaults to ``App.Meta.label``).
        """

        ignore_deprecation_warnings = False
        """Disable deprecation warnings from being logged by Cement."""

        template_module = None
        """
        A python package (dotted import path) where template files can be
        loaded from.  This is generally something like ``myapp.templates``
        where a plugin file would live at ``myapp/templates/mytemplate.txt``.
        Templates are first loaded from ``App.Meta.template_dirs``, and
        and secondly from ``App.Meta.template_module``.  The
        ``template_dirs`` setting has presedence.
        """

        template_dirs = None
        """
        A list of directory paths where template files can be loaded
        from (appended to the builtin list of directories defined by Cement).

        Note: Though ``App.Meta.template_dirs`` defaults to ``None``,
        Cement will populate this with a default list based on
        ``App.Meta.label``.  This will equate to:

        .. code-block:: python

            [
                '~/.myapp/templates',
                '~/.config/myapp/templates',
                '/usr/lib/myapp/templates',
            ]


        Templates are attempted to be loaded in order, and will stop loading
        once a template is successfully loaded from a directory.
        """

        template_dir = None
        """
        A directory path where template files can be loaded from.  By default,
        this setting is also overridden by the
        ``myapp.template_dir`` config setting parsed in any of the
        application configuration files .

        If set, this item will be **prepended** to
        ``App.Meta.template_dirs`` (giving it precedence over other
        ``template_dirs``.
        """

        framework_logging = True
        """
        Whether or not to enable Cement framework logging.  This is separate
        from the application log, and is generally used for debugging issues
        with the framework and/or extensions primarily in development.

        This option is overridden by the environment variable
        `CEMENT_FRAMEWORK_LOGGING`.  Therefore, if in production you do not
        want the Cement framework log enabled, you can set this option to
        ``False`` but override it in your environment by doing something like
        ``export CEMENT_FRAMEWORK_LOGGING=1`` in your shell whenever you need
        it enabled.
        """

        define_hooks = []
        """
        List of hook definitions (labels).  Will be passed to
        ``self.hook.define(<hook_label>)``.  Must be a list of strings.

        I.e. ``['my_custom_hook', 'some_other_hook']``
        """

        hooks = []
        """
        List of hooks to register when the app is created.  Will be passed to
        ``self.hook.register(<hook_label>, <hook_func>)``.  Must be a list of
        tuples in the form of ``(<hook_label>, <hook_func>)``.

        I.e. ``[('post_argument_parsing', my_hook_func)]``.
        """

        core_interfaces = [
            extension.ExtensionInterface,
            log.LogInterface,
            config.ConfigInterface,
            mail.MailInterface,
            plugin.PluginInterface,
            output.OutputInterface,
            template.TemplateInterface,
            arg.ArgumentInterface,
            controller.ControllerInterface,
            cache.CacheInterface,
        ]
        """
        List of core interfaces to be defined (by the framework).  You should
        not modify this unless you really know what you're doing... instead,
        you probably want to add your own interfaces to
        ``App.Meta.interfaces``.
        """

        interfaces = []
        """
        List of interfaces to be defined.  Must be a list of
        uninstantiated interface base classes.

        I.e. ``[MyCustomInterface, SomeOtherInterface]``
        """

        handlers = []
        """
        List of handler classes to register.  Will be passed to
        ``handler.register(<handler_class>)``.  Must be a list of
        uninstantiated handler classes.

        I.e. ``[MyCustomHandler, SomeOtherHandler]``
        """

        alternative_module_mapping = {}
        """
        EXPERIMENTAL FEATURE: This is an experimental feature added in Cement
        2.9.x and may or may not be removed in future versions of Cement.

        Dictionary of alternative, **drop-in** replacement modules to use
        selectively throughout the application, framework, or
        extensions.  Developers can optionally use the
        ``App.__import__()`` method to import simple modules, and if
        that module exists in this mapping it will import the alternative
        library in it's place.

        This is a low-level feature, and may not produce the results you are
        expecting.  It's purpose is to allow the developer to replace specific
        modules at a high level.  Example: For an application wanting to use
        ``ujson`` in place of ``json``, the developer could set the following:

        .. code-block:: python

            alternative_module_mapping = {
                'json' : 'ujson',
            }

        In the app, you would then load ``json`` as:

        .. code-block:: python

            _json = app.__import__('json')
            _json.dumps(data)


        Obviously, the replacement module **must be** a drop-in replace and
        function the same.
        """

        core_system_config_dirs = [
            join('/', 'etc', '{label}', 'ext.d'),
            join('/', 'etc', '{label}', 'plugins.d'),
        ]
        """
        List of builtin system level configuration directories to scan for
        config files.
        """

        core_user_config_dirs = [
            join('{home_dir}', '.config', '{label}', 'ext.d'),
            join('{home_dir}', '.config', '{label}', 'plugins.d'),
            join('{home_dir}', '.{label}', 'config', 'ext.d'),
            join('{home_dir}', '.{label}', 'config', 'plugins.d'),
        ]
        """
        List of builtin user level configuration directories to scan for
        config files.
        """

        core_system_config_files = [
            join('/', 'etc', '{label}', '{label}{suffix}'),
        ]
        """
        List of builtin system level configuration files.
        """

        core_user_config_files = [
            join('{home_dir}', '.config', '{label}', '{label}{suffix}'),
            join('{home_dir}', '.{label}', 'config', '{label}{suffix}'),
            join('{home_dir}', '.{label}{suffix}'),
        ]
        """
        List of builtin user level configuration files.
        """

        core_system_template_dirs = [
            '/usr/lib/{label}/templates',
        ]
        """
        List of builtin system level directories to scan for templates.
        """

        core_user_template_dirs = [
            join('{home_dir}', '.config', '{label}', 'templates'),
            join('{home_dir}', '.{label}', 'templates'),
        ]
        """
        List of builtin user level template directories to scan for templates.
        """

        core_system_plugin_dirs = [
            '/usr/lib/{label}/plugins',
        ]
        """
        List of builtin system level directories to scan for plugins.
        """

        core_user_plugin_dirs = [
            join('{home_dir}', '.config', '{label}', 'plugins'),
            join('{home_dir}', '.{label}', 'plugins'),
        ]
        """
        List of builtin user level directories to scan for plugins.
        """

    def __init__(self, label=None, **kw):
        super(App, self).__init__(**kw)

        # disable framework logging?
        if 'CEMENT_FRAMEWORK_LOGGING' not in os.environ.keys():
            if self._meta.framework_logging is True:
                os.environ['CEMENT_FRAMEWORK_LOGGING'] = '1'
            else:
                os.environ['CEMENT_FRAMEWORK_LOGGING'] = '0'

        # for convenience we translate this to _meta
        if label:
            self._meta.label = label

        self._validate_label()
        self._loaded_bootstrap = None
        self._parsed_args = None
        self._last_rendered = None
        self._extended_members = []
        self.__saved_stdout__ = None
        self.__saved_stderr__ = None
        self.__retry_hooks__ = []
        self.handler = None
        self.interface = None
        self.hook = None
        self.exit_code = 0
        self.ext = None
        self.config = None
        self.log = None
        self.plugin = None
        self.args = None
        self.output = None
        self.controller = None
        self.cache = None
        self.mail = None

        # setup argv... this has to happen before lay_cement()
        if self._meta.argv is None:
            self._meta.argv = list(sys.argv[1:])

        # hack for command line debug/quiet options
        if self._meta.debug_argument_options is not None:
            if any(x in self.argv for x in self._meta.debug_argument_options):
                self._meta.debug = True

        if self._meta.quiet_argument_options is not None:
            if any(x in self.argv for x in self._meta.quiet_argument_options):
                self._meta.quiet = True
                self._suppress_output()

        self._lay_cement()

    @property
    def label(self):
        return self._meta.label

    @property
    def debug(self):
        """
        Returns boolean based on whether the ``debug`` option was passed at
        command line or set via the application's configuration file.

        :returns: boolean
        """
        return self._meta.debug

    @property
    def argv(self):
        """The arguments list that will be used when self.run() is called."""
        return self._meta.argv

    def extend(self, member_name, member_object):
        """
        Extend the ``App()`` object with additional functions/classes such
        as ``app.my_custom_function()``, etc.  It provides an interface for
        extensions to provide functionality that travel along with the
        application object.

        Args:
            member_name (str): The name to attach the object to.
            member_object: The function or class object to attach to
            ``App()``.

        Raises:
            cement.core.exc.FrameworkError: If ``App().member_name`` already
                exists.

        """
        if hasattr(self, member_name):
            raise exc.FrameworkError("App member '%s' already exists!" %
                                     member_name)
        LOG.debug("extending appication with '.%s' (%s)" %
                  (member_name, member_object))
        setattr(self, member_name, member_object)
        if member_name not in self._extended_members:
            self._extended_members.append(member_name)

    def _validate_label(self):
        if not self._meta.label:
            raise exc.FrameworkError("Application name missing.")

        # validate the name is ok
        ok = ['_', '-']
        for char in self._meta.label:
            if char in ok:
                continue

            if not char.isalnum():
                raise exc.FrameworkError(
                    "App label can only contain alpha-numeric, dashes, " +
                    "or underscores."
                )

    def setup(self):
        """
        This function wraps all ``_setup`` actons in one call.  It is called
        before ``self.run()``, allowing the application to be setup but not
        executed (possibly letting the developer perform other actions
        before full execution).

        All handlers should be instantiated and callable after setup is
        complete.

        """
        LOG.debug("now setting up the '%s' application" % self._meta.label)

        if self._meta.bootstrap is not None:
            LOG.debug("importing bootstrap code from %s" %
                      self._meta.bootstrap)

            if self._meta.bootstrap not in sys.modules \
                    or self._loaded_bootstrap is None:
                __import__(self._meta.bootstrap, globals(), locals(), [], 0)
                if hasattr(sys.modules[self._meta.bootstrap], 'load'):
                    sys.modules[self._meta.bootstrap].load(self)

                self._loaded_bootstrap = sys.modules[self._meta.bootstrap]
            else:
                reload_module(self._loaded_bootstrap)

        for res in self.hook.run('pre_setup', self):
            pass

        self._setup_extension_handler()
        self._setup_signals()
        self._setup_config_handler()
        self._setup_mail_handler()
        self._setup_cache_handler()
        self._setup_log_handler()
        self._setup_plugin_handler()
        self._setup_arg_handler()
        self._setup_output_handler()
        self._setup_template_handler()
        self._setup_controllers()

        for hook_spec in self.__retry_hooks__:
            self.hook.register(*hook_spec)

        for res in self.hook.run('post_setup', self):
            pass

    def run(self):
        """
        This function wraps everything together (after ``self._setup()`` is
        called) to run the application.

        Returns:
            unknown: The result of the executed controller function if
            a base controller is set and a controller function is called,
            otherwise ``None`` if no controller dispatched or no controller
            function was called.

        """
        return_val = None

        LOG.debug('running pre_run hook')
        for res in self.hook.run('pre_run', self):
            pass

        # If controller exists, then dispatch it
        if self.controller:
            return_val = self.controller._dispatch()
        else:
            self._parse_args()  # pragma: nocover

        LOG.debug('running post_run hook')
        for res in self.hook.run('post_run', self):
            pass

        return return_val

    def run_forever(self, interval=1, tb=True):
        """
        This function wraps ``self.run()`` with an endless while loop.  If any
        exception is encountered it will be logged and then the application
        will be reloaded.

        Args:
            interval (int): The number of seconds to sleep before reloading the
                the appliction.
            tb (bool): Whether or not to print traceback if exception occurs.

        """

        if tb is True:
            import traceback

        while True:
            LOG.debug('inside run_forever() eternal loop')
            try:
                self.run()
            except Exception as e:
                self.log.fatal('Caught Exception: %s' % e)

                if tb is True:
                    exc_type, exc_value, exc_traceback = sys.exc_info()
                    traceback.print_exception(
                        exc_type, exc_value, exc_traceback, limit=2,
                        file=sys.stdout
                    )
            sleep(interval)
            self.reload()

    def reload(self):
        """
        This function is useful for reloading a running applications, for
        example to reload configuration settings, etc.
        """
        LOG.debug('reloading the %s application' % self._meta.label)
        self._unlay_cement()
        self._lay_cement()
        self.setup()

    def _unlay_cement(self):
        for member in self._extended_members:
            delattr(self, member)
        self._extended_members = []
        self.handler.__handlers__ = {}
        self.hook.__hooks__ = {}

    def close(self, code=None):
        """
        Close the application.  This runs the ``pre_close`` and ``post_close``
        hooks allowing plugins/extensions/etc to cleanup at the end of
        program execution.

        Args:
            code: An exit code to exit with (``int``), if ``None`` is
            passed then exit with whatever ``self.exit_code`` is currently set
            to.  Note: ``sys.exit()`` will only be called if
            ``App.Meta.exit_on_close==True``.
        """
        for res in self.hook.run('pre_close', self):
            pass

        LOG.debug("closing the %s application" % self._meta.label)

        # in theory, this should happen last-last... but at that point `self`
        # would be kind of busted after _unlay_cement() is run.
        for res in self.hook.run('post_close', self):
            pass

        self._unlay_cement()

        if code is not None:
            assert type(code) == int, \
                "Invalid exit status code (must be integer)"
            self.exit_code = code

        if self._meta.exit_on_close is True:
            sys.exit(self.exit_code)

    def render(self, data, template=None, out=sys.stdout, handler=None, **kw):
        """
        This is a simple wrapper around ``self.output.render()`` which simply
        returns an empty string if no output handler is defined.

        Args:
            data (dict): The data dictionary to render.

        Keyword Args:
            template (str): The template to render to (note that some
                output handlers do not use templates).
            out: A file like object (i.e. ``sys.stdout``, or actual file).
                Set to ``None`` if no output is desired (just render and
                return).
            handler: The output handler to use to render.  Defaults to
                ``App.Meta.output_handler``.

        Other Parameters:
            kw (dict): Additional keyword arguments will be passed to the
                output handler when calling ``self.output.render()``.

        """
        for res in self.hook.run('pre_render', self, data):
            if not type(res) is dict:
                LOG.debug("pre_render hook did not return a dict().")
            else:
                data = res

        kw['template'] = template

        if handler is not None:
            oh = self.handler.resolve('output', handler)
            oh._setup(self)
        else:
            oh = self.output

        if oh is None:
            LOG.debug('render() called, but no output handler defined.')
            out_text = ''
        else:
            out_text = oh.render(data, **kw)

        for res in self.hook.run('post_render', self, out_text):
            if not type(res) is str:
                LOG.debug('post_render hook did not return a str()')
            else:
                out_text = str(res)

        if out is not None and not hasattr(out, 'write'):
            raise TypeError("Argument 'out' must be a 'file' like object")
        elif out is not None and out_text is None:
            LOG.debug('render() called but output text is None')
        elif out:
            out.write(out_text)

        self._last_rendered = (data, out_text)
        return out_text

    @property
    def last_rendered(self):
        """
        Return the ``(data, output_text)`` tuple of the last time
        ``self.render()`` was called.

        Returns:
            tuple: ``(data, output_text)``

        """
        return self._last_rendered

    @property
    def pargs(self):
        """
        Returns the ``parsed_args`` object as returned by
        ``self.args.parse()``.
        """
        return self._parsed_args

    def add_arg(self, *args, **kw):
        """A shortcut for ``self.args.add_argument``."""
        self.args.add_argument(*args, **kw)

    def _suppress_output(self):
        if self._meta.debug is True:
            LOG.debug('not suppressing console output because of debug mode')
            return

        LOG.debug('suppressing all console output')
        self.__saved_stdout__ = sys.stdout
        self.__saved_stderr__ = sys.stderr
        sys.stdout = open(os.devnull, 'w')
        sys.stderr = open(os.devnull, 'w')

    def _unsuppress_output(self):
        LOG.debug('unsuppressing all console output')
        sys.stdout = self.__saved_stdout__
        sys.stderr = self.__saved_stderr__

    def _lay_cement(self):
        """Initialize the framework."""
        LOG.debug("laying cement for the '%s' application" %
                  self._meta.label)

        self.interface = InterfaceManager(self)
        self.handler = HandlerManager(self)
        self.hook = HookManager(self)

        # define framework hooks
        self.hook.define('pre_setup')
        self.hook.define('post_setup')
        self.hook.define('pre_run')
        self.hook.define('post_run')
        self.hook.define('pre_argument_parsing')
        self.hook.define('post_argument_parsing')
        self.hook.define('pre_close')
        self.hook.define('post_close')
        self.hook.define('signal')
        self.hook.define('pre_render')
        self.hook.define('post_render')

        # define application hooks from meta
        for label in self._meta.define_hooks:
            self.hook.define(label)

        # register some built-in framework hooks
        self.hook.register('post_setup', add_handler_override_options,
                           weight=-99)
        self.hook.register('post_argument_parsing',
                           handler_override, weight=-99)

        # register application hooks from meta.  the hooks listed in
        # App.Meta.hooks are registered here, so obviously can not be
        # for any hooks other than the builtin framework hooks that we just
        # defined here (above).  Anything that we couldn't register here
        # will be retried after setup
        self.__retry_hooks__ = []
        for hook_spec in self._meta.hooks:
            if not self.hook.defined(hook_spec[0]):
                LOG.debug('hook %s not defined, will retry after setup' %
                          hook_spec[0])
                self.__retry_hooks__.append(hook_spec)
            else:
                self.hook.register(*hook_spec)

        # define interfaces
        for i in self._meta.core_interfaces:
            self.interface.define(i)

        for i in self._meta.interfaces:
            self.interface.define(i)

        # extension handler is the only thing that can't be loaded... as,
        # well, an extension.  ;)
        self.handler.register(extension.ExtensionHandler)

        # register application handlers
        for handler_class in self._meta.handlers:
            self.handler.register(handler_class)

    def _parse_args(self):
        for res in self.hook.run('pre_argument_parsing', self):
            pass

        self._parsed_args = self.args.parse(self.argv)

        for res in self.hook.run('post_argument_parsing', self):
            pass

    def catch_signal(self, signum):
        """
        Add ``signum`` to the list of signals to catch and handle by Cement.

        Args:
            signum (int): The signal number to catch.  See Python
                :py:mod:`signal library <python:signal>`.
        """

        LOG.debug("adding signal handler %s for signal %s" % (
            self._meta.signal_handler, signum)
        )
        signal.signal(signum, self._meta.signal_handler)

    def _setup_signals(self):
        if self._meta.catch_signals is None:
            LOG.debug("catch_signals=None... not handling any signals")
            return

        for signum in self._meta.catch_signals:
            self.catch_signal(signum)

    def _resolve_handler(self, handler_type, handler_def, raise_error=True):
        # meta_defaults = {}
        # if type(handler_def) == str:
        #     _meta_label = "%s.%s" % (handler_type, handler_def)
        #     meta_defaults = self._meta.meta_defaults.get(_meta_label, {})
        # elif hasattr(handler_def, 'Meta'):
        #     _meta_label = "%s.%s" % (handler_type, handler_def.Meta.label)
        #     meta_defaults = self._meta.meta_defaults.get(_meta_label, {})

        han = self.handler.resolve(handler_type,
                                   handler_def,
                                   raise_error=raise_error,
                                   setup=True)
        return han

    def _setup_extension_handler(self):
        LOG.debug("setting up %s.extension handler" % self._meta.label)
        self.ext = self._resolve_handler('extension',
                                         self._meta.extension_handler)
        self.ext.load_extensions(self._meta.core_extensions)
        self.ext.load_extensions(self._meta.extensions)

    def _find_config_files(self, path):
        found_files = []
        if not os.path.isdir(path):
            return []
        files = os.listdir(path)
        files.sort()
        for f in files:
            if f.endswith(self._meta.config_file_suffix):
                found_files.append(fs.join(path, f))
        return found_files

    def _setup_config_handler(self):
        LOG.debug("setting up %s.config handler" % self._meta.label)
        label = self._meta.label
        ext = self._meta.config_file_suffix
        self.config = self._resolve_handler('config',
                                            self._meta.config_handler)
        if self._meta.config_section is None:
            self._meta.config_section = label

        self.config.add_section(self._meta.config_section)

        if self._meta.config_defaults is not None:
            self.config.merge(self._meta.config_defaults)

        if self._meta.config_files is None:
            self._meta.config_files = []

        if self._meta.config_dirs is None:
            self._meta.config_dirs = []

        config_files = []
        config_dirs = []

        template_dict = {
            'label': self._meta.label,
            'suffix': ext,
            'home_dir': fs.HOME_DIR,
        }

        # generate a final list of directories based on precedence (user level
        # paths take precedence).

        for f in self._meta.core_system_config_files:
            f = f.format(**template_dict)
            if f not in config_files:
                config_files.append(f)

        for d in self._meta.core_system_config_dirs:
            d = d.format(**template_dict)
            if d not in config_dirs:
                config_dirs.append(d)
            for f in self._find_config_files(d):
                if f not in config_files:
                    config_files.append(f)

        for f in self._meta.config_files:
            f = f.format(**template_dict)
            if f not in config_files:
                config_files.append(f)

        for d in self._meta.config_dirs:
            d = d.format(**template_dict)
            if d not in config_dirs:
                config_dirs.append(d)
            for f in self._find_config_files(d):
                if f not in config_files:
                    config_files.append(f)

        for f in self._meta.core_user_config_files:
            f = f.format(**template_dict)
            if f not in config_files:
                config_files.append(f)

        for d in self._meta.core_user_config_dirs:
            d = d.format(**template_dict)
            if d not in config_dirs:
                config_dirs.append(d)
            for f in self._find_config_files(d):
                if f not in config_files:
                    config_files.append(f)

        # reset for final lists

        self._meta.config_files = []
        self._meta.config_dirs = []

        for d in config_dirs:
            self.add_config_dir(d)

        for f in config_files:
            self.add_config_file(f)
            self.config.parse_file(f)

        self.validate_config()

        # hack for debug command line option
        if self._meta.debug is True:
            self.config.set(self._meta.config_section, 'debug', True)

        # override select Meta via config
        base_dict = self.config.get_section_dict(self._meta.config_section)

        for key in base_dict:
            if key in self._meta.core_meta_override or \
                    key in self._meta.meta_override:
                # kind of a hack for core_meta_override
                if key in ['debug']:
                    setattr(self._meta, key, is_true(base_dict[key]))
                else:
                    setattr(self._meta, key, base_dict[key])

        # load extensions from configuraton file
        if 'extensions' in self.config.keys(self._meta.config_section):
            exts = self.config.get(label, 'extensions')

            # convert a comma-separated string to a list
            if type(exts) is str:
                ext_list = exts.split(',')

                # clean up extra space if they had it inbetween commas
                ext_list = [x.strip() for x in ext_list]

                # set the new extensions value in the config
                self.config.set(label, 'extensions', ext_list)

            # otherwise, if it's a list
            elif type(exts) is list:
                ext_list = exts

            for ext in ext_list:
                # load the extension
                self.ext.load_extension(ext)

                # add to meta data
                self._meta.extensions.append(ext)

    def _setup_mail_handler(self):
        LOG.debug("setting up %s.mail handler" % self._meta.label)
        self.mail = self._resolve_handler('mail',
                                          self._meta.mail_handler)

    def _setup_log_handler(self):
        LOG.debug("setting up %s.log handler" % self._meta.label)
        self.log = self._resolve_handler('log', self._meta.log_handler)

    def _setup_plugin_handler(self):
        LOG.debug("setting up %s.plugin handler" % self._meta.label)

        # plugin dirs
        if self._meta.plugin_dirs is None:
            self._meta.plugin_dirs = []

        plugin_dirs = []
        template_dict = {
            'label': self._meta.label,
            'home_dir': fs.HOME_DIR,
        }

        # generate a final list of directories based on precedence (user level
        # paths take precedence).

        for d in self._meta.core_system_plugin_dirs:
            d = d.format(**template_dict)
            if d not in plugin_dirs:
                plugin_dirs.append(d)

        for d in self._meta.plugin_dirs:
            d = d.format(**template_dict)
            if d not in plugin_dirs:
                plugin_dirs.append(d)

        if self._meta.plugin_dir is not None:
            d = self._meta.plugin_dir.format(**template_dict)
            plugin_dirs.append(d)

        for d in self._meta.core_user_plugin_dirs:
            d = d.format(**template_dict)
            if d not in plugin_dirs:
                plugin_dirs.append(d)

        # reset our final list

        self._meta.plugin_dirs = []

        # reverse it so that first found takes precedence
        plugin_dirs.reverse()

        for path in plugin_dirs:
            self.add_plugin_dir(path)

        # plugin bootstrap
        if self._meta.plugin_module is None:
            self._meta.plugin_module = '%s.plugins' % self._meta.label

        self.plugin = self._resolve_handler('plugin',
                                            self._meta.plugin_handler)
        self.plugin.load_plugins(self._meta.plugins)
        self.plugin.load_plugins(self.plugin.get_enabled_plugins())

    def _setup_output_handler(self):
        if self._meta.output_handler is None:
            LOG.debug("no output handler defined, skipping.")
            return

        LOG.debug("setting up %s.output handler" % self._meta.label)
        self.output = self._resolve_handler('output',
                                            self._meta.output_handler,
                                            raise_error=False)

    def _setup_template_handler(self):
        if self._meta.template_handler is None:
            LOG.debug("no template handler defined, skipping.")
            return

        label = self._meta.label
        LOG.debug("setting up %s.template handler" % self._meta.label)
        self.template = self._resolve_handler('template',
                                              self._meta.template_handler,
                                              raise_error=False)
        # template module
        if self._meta.template_module is None:
            self._meta.template_module = '%s.templates' % label

        # template dirs
        if self._meta.template_dirs is None:
            self._meta.template_dirs = []

        template_dirs = []
        template_dict = {
            'label': self._meta.label,
            'home_dir': fs.HOME_DIR,
        }

        # generate a final list of directories based on precedence (user level
        # paths take precedence).

        for d in self._meta.core_system_template_dirs:
            d = d.format(**template_dict)
            if d not in template_dirs:
                template_dirs.append(d)

        for d in self._meta.template_dirs:
            d = d.format(**template_dict)
            if d not in template_dirs:
                template_dirs.append(d)

        if self._meta.template_dir is not None:
            d = self._meta.template_dir.format(**template_dict)
            template_dirs.append(d)

        for d in self._meta.core_user_template_dirs:
            d = d.format(**template_dict)
            if d not in template_dirs:
                template_dirs.append(d)

        # reset final list
        self._meta.template_dirs = []

        # reverse so that first found has precedence
        template_dirs.reverse()

        for path in template_dirs:
            self.add_template_dir(path)

    def _setup_cache_handler(self):
        if self._meta.cache_handler is None:
            LOG.debug("no cache handler defined, skipping.")
            return

        LOG.debug("setting up %s.cache handler" % self._meta.label)
        self.cache = self._resolve_handler('cache',
                                           self._meta.cache_handler,
                                           raise_error=False)

    def _setup_arg_handler(self):
        LOG.debug("setting up %s.arg handler" % self._meta.label)
        self.args = self._resolve_handler('argument',
                                          self._meta.argument_handler)
        self.args.prog = self._meta.label

        if self._meta.debug_argument_options is not None:
            self.args.add_argument(*self._meta.debug_argument_options,
                                   dest='debug',
                                   action='store_true',
                                   help=self._meta.debug_argument_help)
        if self._meta.quiet_argument_options is not None:
            self.args.add_argument(*self._meta.quiet_argument_options,
                                   dest='suppress_output',
                                   action='store_true',
                                   help=self._meta.quiet_argument_help)

        # merge handler override meta data
        if self._meta.handler_override_options is not None:
            # merge the core handler override options with developer defined
            # options
            core = self._meta.core_handler_override_options.copy()
            dev = self._meta.handler_override_options.copy()
            core.update(dev)

            self._meta.handler_override_options = core

    def _setup_controllers(self):
        LOG.debug("setting up application controllers")

        if self.handler.registered('controller', 'base'):
            self.controller = self._resolve_handler('controller', 'base')

        else:
            class DefaultBaseController(Controller):
                class Meta:
                    label = 'base'

                def _default(self):
                    # don't enforce anything cause developer might not be
                    # using controllers... if they are, they should define
                    # a base controller.
                    pass

            self.handler.register(DefaultBaseController)
            self.controller = self._resolve_handler('controller', 'base')

    def validate_config(self):
        """
        Validate application config settings.

        Example:

        .. code-block:: python

            import os
            from cement import App

            class MyApp(App):
                class Meta:
                    label = 'myapp'

                def validate_config(self):
                    super(MyApp, self).validate_config()

                    # test that the log file directory exist, if not create it
                    logdir = os.path.dirname(self.config.get('log', 'file'))

                    if not os.path.exists(logdir):
                        os.makedirs(logdir)

        """
        pass

    def add_config_dir(self, path):
        """
        Append a directory ``path`` to the list of directories to parse for
        config files.

        Args:
            path (str): Directory path that contains config files.

        Example:

        .. code-block:: python

            app.add_config_dir('/path/to/my/config/')

        """
        path = fs.abspath(path)
        if path not in self._meta.config_dirs:
            self._meta.config_dirs.append(path)

    def add_config_file(self, path):
        """
        Append a file ``path`` to the list of configuration files to parse.

        Args:
            path (str): Configuration file path..

        Example:

        .. code-block:: python

            app.add_config_file('/path/to/my/config/file')

        """
        path = fs.abspath(path)
        if path not in self._meta.config_files:
            self._meta.config_files.append(path)

    def add_plugin_dir(self, path):
        """
        Append a directory ``path`` to the list of directories to scan for
        plugins.

        Args:
            path (str): Directory path that contains plugin files.

        Example:

        .. code-block:: python

            app.add_plugin_dir('/path/to/my/plugins')

        """
        path = fs.abspath(path)
        if path not in self._meta.plugin_dirs:
            self._meta.plugin_dirs.append(path)

    def add_template_dir(self, path):
        """
        Append a directory ``path`` to the list of template directories to
        parse for templates.

        Args:
            path (str): Directory path that contains template files.

        Example:

        .. code-block:: python

            app.add_template_dir('/path/to/my/templates')

        """
        path = fs.abspath(path)
        if path not in self._meta.template_dirs:
            self._meta.template_dirs.append(path)

    def remove_template_dir(self, path):
        """
        Remove a directory ``path`` from the list of template directories to
        parse for templates.

        Args:
            path (str): Directory path that contains template files.

        Example:

            .. code-block:: python

                app.remove_template_dir('/path/to/my/templates')

        """
        path = fs.abspath(path)
        if path in self._meta.template_dirs:
            self._meta.template_dirs.remove(path)

    def __import__(self, obj, from_module=None):
        # EXPERIMENTAL == UNDOCUMENTED
        mapping = self._meta.alternative_module_mapping

        if from_module is not None:
            _from = mapping.get(from_module, from_module)
            _loaded = __import__(_from, globals(), locals(), [obj], 0)
            return getattr(_loaded, obj)
        else:
            obj = mapping.get(obj, obj)
            _loaded = __import__(obj, globals(), locals(), [], 0)

        return _loaded

    def __enter__(self):
        self.setup()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        # only close the app if there are no unhandled exceptions
        if exc_type is None:
            self.close()


class TestApp(App):

    """
    App subclass useful for testing.

    """

    # tells pytest to not consider this a class for testing
    __test__ = False

    class Meta:
        label = "app-%s" % misc.rando()[:12]
        argv = []
        core_system_config_files = []
        core_user_config_files = []
        config_files = []
        core_system_config_dirs = []
        core_user_config_dirs = []
        config_dirs = []
        core_system_template_dirs = []
        core_user_template_dirs = []
        core_system_plugin_dirs = []
        core_user_plugin_dirs = []
        plugin_dirs = []
        exit_on_close = False



"""
Cement print extension module.
"""

from ..core import output
from ..utils.misc import minimal_logger

LOG = minimal_logger(__name__)


def extend_print(app):
    def _print(text):
        app.render({'out': text}, handler='print')
    app.extend('print', _print)


class PrintOutputHandler(output.OutputHandler):

    """
    This class implements the :ref:`Output <cement.core.output>` Handler
    interface.  It takes a dict and only prints out the ``out`` key. It is
    primarily used by the ``app.print()`` extended function in order to replace
    ``print()`` so that framework features like ``pre_render`` and
    ``post_render`` hooks are honored. Please see the developer documentation
    on :cement:`Output Handling <dev/output>`.

    """
    class Meta:

        """Handler meta-data"""

        label = 'print'
        """The string identifier of this handler."""

        #: Whether or not to include ``json`` as an available choice
        #: to override the ``output_handler`` via command line options.
        overridable = False

    def render(self, data_dict, template=None, **kw):
        """
        Take a data dictionary and render only the ``out`` key as text output.
        Note that the template option is received here per the interface,
        however this handler just ignores it.

        Args:
            data_dict (dict): The data dictionary to render.

        Keyword Args:
            template: This option is completely ignored.

        Returns:
            str: A text string.

        """
        if 'out' in data_dict.keys():
            LOG.debug("rendering content as text via %s" % self.__module__)
            return data_dict['out'] + '\n'
        else:
            LOG.debug("no 'out' key found in data dict. "
                      "not rendering content via %s" % self.__module__)
            return None


class PrintDictOutputHandler(output.OutputHandler):

    """
    This class implements the :ref:`Output <cement.core.output>` Handler
    interface.  It is intended primarily for development where printing out a
    string reprisentation of the data dictionary would be useful.  Please see
    the developer documentation on :cement:`Output Handling <dev/output>`.

    """
    class Meta:

        """Handler meta-data"""

        label = 'print_dict'
        """The string identifier of this handler."""

        #: Whether or not to include ``json`` as an available choice
        #: to override the ``output_handler`` via command line options.
        overridable = False

    def render(self, data_dict, template=None, **kw):
        """
        Take a data dictionary and render it as text output.  Note that the
        template option is received here per the interface, however this
        handler just ignores it.

        Args:
            data_dict (dict): The data dictionary to render.

        Keyword Args:
            template: This option is completely ignored.

        Returns:
            str: A text string.

        """
        LOG.debug("rendering content as text via %s" % self.__module__)
        out = ''
        for key, val in data_dict.items():
            out = out + '%s: %s\n' % (key, val)

        return out


def load(app):
    app.handler.register(PrintDictOutputHandler)
    app.handler.register(PrintOutputHandler)
    app.hook.register('pre_argument_parsing', extend_print)



"""Misc utilities."""

import os
import sys
import logging
import hashlib
from textwrap import TextWrapper
from random import random


def rando(salt=None):
    """
    Generate a random MD5 hash for whatever purpose.  Useful for testing
    or any other time that something random is required.

    Args:
        salt (str): Optional ``salt``, if ``None`` then ``random.random()``
            is used.

    Returns:
        str: Random MD5 hash

    Example:

        .. code-block:: python

            from cement.utils.misc import rando

            rando('dAhn49amvnah3m')

    """

    if salt is None:
        salt = random()

    return hashlib.md5(str(salt).encode()).hexdigest()


class MinimalLogger(object):

    def __init__(self, namespace, debug, *args, **kw):
        self.namespace = namespace
        self.backend = logging.getLogger(namespace)
        formatter = logging.Formatter(
            "%(asctime)s (%(levelname)s) %(namespace)s : %(message)s"
        )
        console = logging.StreamHandler()
        console.setFormatter(formatter)
        console.setLevel(logging.INFO)
        self.backend.setLevel(logging.INFO)

        # FIX ME: really don't want to hard check sys.argv like this but
        # can't figure any better way get logging started (only for debug)
        # before the app logging is setup.
        if '--debug' in sys.argv or debug:
            console.setLevel(logging.DEBUG)
            self.backend.setLevel(logging.DEBUG)

        self.backend.addHandler(console)

    def _get_logging_kwargs(self, namespace, **kw):
        if not namespace:
            namespace = self.namespace

        if 'extra' in kw.keys() and 'namespace' in kw['extra'].keys():
            pass
        elif 'extra' in kw.keys() and 'namespace' not in kw['extra'].keys():
            kw['extra']['namespace'] = namespace
        else:
            kw['extra'] = dict(namespace=namespace)

        return kw

    @property
    def logging_is_enabled(self):
        if 'CEMENT_FRAMEWORK_LOGGING' in os.environ.keys():
            if is_true(os.environ['CEMENT_FRAMEWORK_LOGGING']):
                res = True
            else:
                res = False
        else:
            res = True  # pragma: nocover

        return res

    def info(self, msg, namespace=None, **kw):
        if self.logging_is_enabled:
            kwargs = self._get_logging_kwargs(namespace, **kw)
            self.backend.info(msg, **kwargs)

    def warning(self, msg, namespace=None, **kw):
        if self.logging_is_enabled:
            kwargs = self._get_logging_kwargs(namespace, **kw)
            self.backend.warning(msg, **kwargs)

    def error(self, msg, namespace=None, **kw):
        if self.logging_is_enabled:
            kwargs = self._get_logging_kwargs(namespace, **kw)
            self.backend.error(msg, **kwargs)

    def fatal(self, msg, namespace=None, **kw):
        if self.logging_is_enabled:
            kwargs = self._get_logging_kwargs(namespace, **kw)
            self.backend.fatal(msg, **kwargs)

    def debug(self, msg, namespace=None, **kw):
        if self.logging_is_enabled:
            kwargs = self._get_logging_kwargs(namespace, **kw)
            self.backend.debug(msg, **kwargs)


def init_defaults(*sections):
    """
    Returns a standard dictionary object to use for application defaults.
    If sections are given, it will create a nested dict for each section name.
    This is sometimes more useful, or cleaner than creating an entire dict set
    (often used in testing).

    Args:
        sections: Section keys to create nested dictionaries for.

    Returns:
        dict: Dictionary of nested dictionaries (sections)

    Example:

        .. code-block:: python

            from cement import App, init_defaults

            defaults = init_defaults('myapp', 'section2', 'section3')
            defaults['myapp']['debug'] = False
            defaults['section2']['foo'] = 'bar
            defaults['section3']['foo2'] = 'bar2'

            app = App('myapp', config_defaults=defaults)

    """
    defaults = dict()
    for section in sections:
        defaults[section] = dict()
    return defaults


def minimal_logger(namespace, debug=False):
    """
    Setup just enough for cement to be able to do debug logging.  This is the
    logger used by the Cement framework, which is setup and accessed before
    the application is functional (and more importantly before the
    applications log handler is usable).

    Args:
        namespace (str): The logging namespace.  This is generally
            ``__name__`` or anything you want.

    Keyword Args:
        debug (bool): Toggle debug output.

    Returns:
        object: A Logger object

    Example:

        .. code-block:: python

            from cement.utils.misc import minimal_logger
            LOG = minimal_logger('cement')
            LOG.debug('This is a debug message')

    """
    return MinimalLogger(namespace, debug)


def is_true(item):
    """
    Given a value, determine if it is one of
    ``[True, 'true', 'yes', 'on', '1', 1,]`` (note: strings are converted to
    lowercase before comparison).

    Args:
        item: The item to convert to a boolean.

    Returns:
        bool: ``True`` if ``item`` equates to a true-ish value, ``False``
            otherwise

    """
    if isinstance(item, str) and item.lower() in ['true', 'yes', 'on', '1']:
        return True
    elif isinstance(item, bool) and item is True:
        return True
    elif isinstance(item, int) and item == 1:
        return True
    else:
        return False


def wrap(text, width=77, indent='', long_words=False, hyphens=False):
    """
    Wrap text for cleaner output (this is a simple wrapper around
    ``textwrap.TextWrapper`` in the standard library).

    Args:
        text (str): The text to wrap

    Keyword Arguments:
        width (int): The max width of a line before breaking
        indent (str): String to prefix subsequent lines after breaking
        long_words (bool): Whether or not to break on long words
        hyphens (bool): Whether or not to break on hyphens

    Returns:
        str: The wrapped string

    """

    types = [str]
    if type(text) not in types:
        raise TypeError("Argument `text` must be one of [str, unicode].")

    wrapper = TextWrapper(subsequent_indent=indent, width=width,
                          break_long_words=long_words,
                          break_on_hyphens=hyphens)
    return wrapper.fill(text)



"""
PyTest Fixtures.
"""

import pytest
from cement import fs

@pytest.fixture(scope="function")
def tmp(request):
    """
    Create a `tmp` object that geneates a unique temporary directory, and file
    for each test function that requires it.
    """
    t = fs.Tmp()
    yield t
    t.remove()



"""
Cement memcached extension module.
"""

import pylibmc
from ..core import cache
from ..utils.misc import minimal_logger

LOG = minimal_logger(__name__)


class MemcachedCacheHandler(cache.CacheHandler):

    """
    This class implements the :ref:`Cache <cement.core.cache>` Handler
    interface.  It provides a caching interface using the
    `pylibmc <http://sendapatch.se/projects/pylibmc/>`_ library.

    **Note** This extension has an external dependency on ``pylibmc``.  You
    must include ``pylibmc`` in your applications dependencies as Cement
    explicitly does *not* include external dependencies for optional
    extensions.
    """

    class Meta:

        """Handler meta-data."""

        label = 'memcached'
        config_defaults = dict(
            hosts=['127.0.0.1'],
            expire_time=0,
        )

    def __init__(self, *args, **kw):
        super(MemcachedCacheHandler, self).__init__(*args, **kw)
        self.mc = None

    def _setup(self, *args, **kw):
        super(MemcachedCacheHandler, self)._setup(*args, **kw)
        self._fix_hosts()
        self.mc = pylibmc.Client(self._config('hosts'))

    def _fix_hosts(self):
        """
        Useful to fix up the hosts configuration (i.e. convert a
        comma-separated string into a list).  This function does not return
        anything, however it is expected to set the `hosts` value of the
        ``[cache.memcached]`` section (which is what this extension reads for
        it's host configution).

        :returns: ``None``

        """
        hosts = self._config('hosts')
        fixed_hosts = []

        if type(hosts) == str:
            parts = hosts.split(',')
            for part in parts:
                fixed_hosts.append(part.strip())
        elif type(hosts) == list:
            fixed_hosts = hosts
        self.app.config.set(self._meta.config_section, 'hosts', fixed_hosts)

    def get(self, key, fallback=None, **kw):
        """
        Get a value from the cache.  Any additional keyword arguments will be
        passed directly to `pylibmc` get function.

        Args:
            key (str): The key of the item in the cache to get.

        Keyword Args:
            fallback: The value to return if the item is not found in the
                cache.

        Returns:
            unknown: The value of the item in the cache, or the ``fallback``
            value.

        """
        LOG.debug("getting cache value using key '%s'" % key)
        res = self.mc.get(key, **kw)
        if res is None:
            return fallback
        else:
            return res

    def _config(self, key):
        """
        This is a simple wrapper, and is equivalent to:
        ``self.app.config.get('cache.memcached', <key>)``.

        Args:
            key (str): The key to get a config value from the 'cache.memcached'
                config section.

        Returns:
            unknown: The value of the given key.

        """
        return self.app.config.get(self._meta.config_section, key)

    def set(self, key, value, time=None, **kw):
        """
        Set a value in the cache for the given ``key``.  Any additional
        keyword arguments will be passed directly to the `pylibmc` set
        function.

        Args:
            key (str): The key of the item in the cache to set.
            value: The value of the item to set.

        Keyword Arguments:
            time (int): The expiration time (in seconds) to keep the item
                cached.  Defaults to `expire_time` as defined in the
                applications configuration.

        """
        if time is None:
            time = int(self._config('expire_time'))

        self.mc.set(key, value, time=time, **kw)

    def delete(self, key, **kw):
        """
        Delete an item from the cache for the given ``key``.  Any additional
        keyword arguments will be passed directly to the `pylibmc` delete
        function.

        Args:
            key (str): The key to delete from the cache.

        """
        self.mc.delete(key, **kw)

    def purge(self, **kw):
        """
        Purge the entire cache, all keys and values will be lost.  Any
        additional keyword arguments will be passed directly to the
        pylibmc ``flush_all()`` function.

        """

        self.mc.flush_all(**kw)


def load(app):
    app.handler.register(MemcachedCacheHandler)




class TodoError(Exception):
    """Generic errors."""

    def __init__(self, msg):
        Exception.__init__(self)
        self.msg = msg

    def __str__(self):
        return self.msg

    def __repr__(self):
        return "<TodoError - %s>" % self.msg



"""
Cement redis extension module.
"""

import redis
from ..core import cache
from ..utils.misc import minimal_logger

LOG = minimal_logger(__name__)


class RedisCacheHandler(cache.CacheHandler):

    """
    This class implements the :ref:`Cache <cement.core.cache>` Handler
    interface.  It provides a caching interface using the
    `redis <http://github.com/andymccurdy/redis-py>`_ library.

    **Note** This extension has an external dependency on ``redis``.  You
    must include ``redis`` in your applications dependencies as Cement
    explicitly does *not* include external dependencies for optional
    extensions.
    """

    class Meta:

        """Handler meta-data."""

        label = 'redis'
        config_defaults = dict(
            host='127.0.0.1',
            port=6379,
            db=0,
            expire_time=0,
        )

    def __init__(self, *args, **kw):
        super(RedisCacheHandler, self).__init__(*args, **kw)
        self.mc = None

    def _setup(self, *args, **kw):
        super(RedisCacheHandler, self)._setup(*args, **kw)
        self.r = redis.StrictRedis(
            host=self._config('host', default='127.0.0.1'),
            port=self._config('port', default=6379),
            db=self._config('db', default=0))

    def _config(self, key, default=None):
        """
        This is a simple wrapper, and is equivalent to:
        ``self.app.config.get('cache.redis', <key>)``.

        Args:
            key (str): The key to get a config value from the ``cache.redis``
                config section.

        Returns:
            unknown: The value of the given key.

        """
        return self.app.config.get(self._meta.config_section, key)

    def get(self, key, fallback=None, **kw):
        """
        Get a value from the cache.  Additional keyword arguments are ignored.

        Args:
            key (str): The key of the item in the cache to get.

        Keyword Args:
            fallback: The value to return if the item is not found in the
                cache.

        Returns:
            unknown: The value of the item in the cache, or the ``fallback``
            value.

        """
        LOG.debug("getting cache value using key '%s'" % key)
        res = self.r.get(key)
        if res is None:
            return fallback
        else:
            return res.decode('utf-8')

    def set(self, key, value, time=None, **kw):
        """
        Set a value in the cache for the given ``key``.  Additional
        keyword arguments are ignored.

        Args:
            key (str): The key of the item in the cache to set.
            value: The value of the item to set.
            time (int): The expiration time (in seconds) to keep the item
                cached. Defaults to ``expire_time`` as defined in the
                applications configuration.

        """
        if time is None:
            time = int(self._config('expire_time'))

        if time == 0:
            self.r.set(key, value)
        else:
            self.r.setex(key, time, value)

    def delete(self, key, **kw):
        """
        Delete an item from the cache for the given ``key``.  Additional
        keyword arguments are ignored.

        Args:
            key (str): The key to delete from the cache.

        """
        self.r.delete(key)

    def purge(self, **kw):
        """
        Purge the entire cache, all keys and values will be lost.  Any
        additional keyword arguments will be passed directly to the
        redis ``flush_all()`` function.

        """
        keys = self.r.keys('*')
        if keys:
            self.r.delete(*keys)


def load(app):
    app.handler.register(RedisCacheHandler)
